"""Implement a callback to log images."""

# =============================================================================
# Copyright 2021 Henrique Morimitsu
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

from pathlib import Path
from typing import Any, Dict, List, Sequence, Tuple, Union

try:
    from neptune.new.types import File as NeptuneFile
except ImportError:
    NeptuneFile = None
import numpy as np
from PIL import Image, ImageDraw, ImageFont
from lightning.pytorch.callbacks import Callback
from lightning.pytorch.loggers.comet import CometLogger
from lightning.pytorch.loggers.neptune import NeptuneLogger
from lightning.pytorch.loggers.tensorboard import TensorBoardLogger
from lightning.pytorch.loggers.wandb import WandbLogger
from lightning.pytorch.trainer.trainer import Trainer
import torch
import torch.nn.functional as F
from torchvision.utils import make_grid

try:
    import wandb
except ImportError:
    wandb = None

from ptlflow.models.base_model.base_model import BaseModel
from ptlflow.utils import flow_utils
from ptlflow.utils.utils import config_logging

config_logging()


class LoggerCallback(Callback):
    """Callback to collect and log images during training and validation.

    For each dataloader, num_images samples will be collected. The samples are collected by trying to retrieve from both
    inputs and outputs tensors whose keys match the values provided in log_keys.

    num_images samples are uniformly sampled from the whole dataloader.
    """

    def __init__(
        self,
        num_images: int = 5,
        image_size: Tuple[int, int] = (200, 400),
        log_keys: Sequence[str] = ("images", "flows", "occs", "mbs", "confs"),
        epe_clip: float = 5.0,
    ) -> None:
        """Initialize LoggerCallback.

        Parameters
        ----------
        num_images : int, default 5
            Number of images to log during one epoch.
        image_size : Tuple[int, int], default (200, 400)
            The size of the stored images.
        log_keys : Sequence[str], default ('images', 'flows', 'occs', 'mbs', 'confs')
            The keys to use to collect the images from the inputs and outputs of the model. If a key is not found, it is
            ignored.
        epe_clip : float, default 5.0
            The maximum EPE value that is shown on EPE image. All EPE values above this will be clipped.
        """
        super().__init__()

        self.num_images = num_images
        self.image_size = image_size
        self.log_keys = log_keys
        self.epe_clip = epe_clip

        self.train_collect_img_idx = []
        self.train_images = {}

        self.val_dataloader_names = []
        self.val_collect_image_idx = {}
        self.val_images = {}

    def log_image(self, title: str, image: torch.Tensor, pl_module: BaseModel) -> None:
        """Log the image in all of the pl_module loggers.

        Note, however, that not all loggers may be able to log images.

        Parameters
        ----------
        title : str
            A title for the image.
        image : torch.Tensor
            The image to log. It must be a 3D tensor CHW (typically C=3).
        pl_module : BaseModel
            An instance of the optical flow model to get the logger from.
        """
        image_npy = image.permute(1, 2, 0).numpy()

        for logger in pl_module.loggers:
            if isinstance(logger, CometLogger):
                logger.experiment.log_image(image_npy, name=title)
            elif isinstance(logger, NeptuneLogger):
                logger.experiment[title].log(NeptuneFile.as_image(image))
            elif isinstance(logger, TensorBoardLogger):
                logger.experiment.add_image(title, image, pl_module.global_step)
            elif isinstance(logger, WandbLogger) and wandb is not None:
                title_wb = title.replace("/", "-")
                image_wb = wandb.Image(image_npy)
                logger.experiment.log({title_wb: image_wb})

    def on_train_batch_end(
        self,
        trainer: Trainer,
        pl_module: BaseModel,
        outputs: Dict[str, torch.Tensor],
        batch: Dict[str, torch.Tensor],
        batch_idx: int,
        **kwargs,
    ) -> None:
        """Store one image to be logged, if the current batch_idx is in the log selection group.

        Parameters
        ----------
        trainer : Trainer
            An instance of the PyTorch Lightning trainer.
        pl_module : BaseModel
            An instance of the optical flow model.
        outputs : Dict[str, torch.Tensor]
            The outputs of the current training batch.
        batch : Dict[str, torch.Tensor]
            The inputs of the current training batch.
        batch_idx : int
            The counter value of the current batch.
        """
        if batch_idx in self.train_collect_img_idx:
            self._append_images(
                self.train_images, pl_module.last_inputs, pl_module.last_predictions
            )

    def on_train_epoch_start(self, trainer: Trainer, pl_module: BaseModel) -> None:
        """Reset the training log params and accumulators.

        Parameters
        ----------
        trainer : Trainer
            An instance of the PyTorch Lightning trainer.
        pl_module : BaseModel
            An instance of the optical flow model.
        """
        self.train_images = {}
        limit_batches = (
            trainer.limit_train_batches
            if trainer.limit_train_batches is not None
            else 1.0
        )
        collect_idx = np.unique(
            np.linspace(
                0,
                self._compute_max_range(
                    trainer.datamodule.train_dataloader_length, limit_batches
                ),
                self.num_images,
                dtype=np.int32,
            )
        )
        self.train_collect_img_idx = collect_idx

    def on_train_epoch_end(
        self, trainer: Trainer, pl_module: BaseModel, **kwargs
    ) -> None:
        """Log the images accumulated during the training.

        Parameters
        ----------
        trainer : Trainer
            An instance of the PyTorch Lightning trainer.
        pl_module : BaseModel
            An instance of the optical flow model.
        outputs : Any
            Outputs of the training epoch.
        """
        img_grid = self._make_image_grid(self.train_images)
        self.log_image("train", img_grid, pl_module)

    def on_validation_batch_end(
        self,
        trainer: Trainer,
        pl_module: BaseModel,
        outputs: Dict[str, torch.Tensor],
        batch: Dict[str, torch.Tensor],
        batch_idx: int,
        dataloader_idx: int = 0,
    ) -> None:
        """Store one image to be logged, if the current batch_idx is in the log selection group.

        Parameters
        ----------
        trainer : Trainer
            An instance of the PyTorch Lightning trainer.
        pl_module : BaseModel
            An instance of the optical flow model.
        outputs : Dict[str, torch.Tensor]
            The outputs of the current validation batch.
        batch : Dict[str, torch.Tensor]
            The inputs of the current validation batch.
        batch_idx : int
            The counter value of the current batch.
        dataloader_idx : int
            The index number of the current dataloader.
        """
        dl_name = self.val_dataloader_names[dataloader_idx]
        if batch_idx in self.val_collect_image_idx[dl_name]:
            self._append_images(
                self.val_images[dl_name],
                pl_module.last_inputs,
                pl_module.last_predictions,
            )

    def on_validation_epoch_start(self, trainer: Trainer, pl_module: BaseModel) -> None:
        """Reset the validation log params and accumulators.

        Parameters
        ----------
        trainer : Trainer
            An instance of the PyTorch Lightning trainer.
        pl_module : BaseModel
            An instance of the optical flow model.
        """
        self.val_dataloader_names = trainer.datamodule.val_dataloader_names
        for dl_name in self.val_dataloader_names:
            self.val_images[dl_name] = {}

        limit_batches = (
            trainer.limit_val_batches if trainer.limit_val_batches is not None else 1.0
        )
        for dname, dlen in zip(
            trainer.datamodule.val_dataloader_names,
            trainer.datamodule.val_dataloader_lengths,
        ):
            collect_idx = np.unique(
                np.linspace(
                    0,
                    self._compute_max_range(dlen, limit_batches),
                    self.num_images,
                    dtype=np.int32,
                )
            )
            self.val_collect_image_idx[dname] = collect_idx

    def on_validation_epoch_end(self, trainer: Trainer, pl_module: BaseModel) -> None:
        """Log the images accumulated during the validation.

        Parameters
        ----------
        trainer : Trainer
            An instance of the PyTorch Lightning trainer.
        pl_module : BaseModel
            An instance of the optical flow model.
        """
        for dl_name, dl_images in self.val_images.items():
            img_grid = self._make_image_grid(dl_images)
            self.log_image(f"val/{dl_name}", img_grid, pl_module)

    def _add_title(self, image: torch.Tensor, img_title: str) -> torch.Tensor:
        """Add a title to an image.

        Parameters
        ----------
        image : torch.Tensor
            The image where the title will be added.
        img_title : str
            The title to be added.

        Returns
        -------
        torch.Tensor
            The input image with the title superposed on it.
        """
        size = min(image.shape[1:3])
        image = (255 * image.permute(1, 2, 0).numpy()).astype(np.uint8)
        image = Image.fromarray(image)

        this_dir = Path(__file__).resolve().parent
        title_font = ImageFont.truetype(
            str(this_dir / "RobotoMono-Regular.ttf"), size // 10
        )

        draw = ImageDraw.Draw(image)
        bb = (
            size // 25,
            size // 25,
            size // 25 + len(img_title) * size // 15,
            size // 25 + size // 8,
        )
        draw.rectangle(bb, fill="black")
        draw.text((size // 20, size // 30), img_title, (237, 230, 211), font=title_font)

        image = np.array(image)
        image = torch.from_numpy(image.transpose(2, 0, 1)).float() / 255

        return image

    def _append_images(  # noqa: C901
        self,
        images: Dict[str, List[torch.Tensor]],
        inputs: Dict[str, torch.Tensor],
        preds: Dict[str, torch.Tensor],
    ) -> None:
        """Append samples to the images accumulator.

        Parameters
        ----------
        images : Dict[str, List[torch.Tensor]]
            The accumulator where the samples will be appended to.
        inputs : Dict[str, torch.Tensor]
            The inputs of the model.
        preds : Dict[str, torch.Tensor]
            The outrputs of the model.
        """
        for k in self.log_keys:
            log_names = []
            log_sources = []
            if k in inputs or (k == "confs" and k in preds):
                log_names.append(f"i_{k}")
                log_sources.append(inputs)
            if k in preds:
                log_names.append(f"o_{k}")
                log_sources.append(preds)

                if k == "flows":
                    log_names.append(f"epe<{self.epe_clip:.1f}")
                    log_sources.append(None)

            for name, source in zip(log_names, log_sources):
                if images.get(name) is None:
                    images[name] = []

                if name == "i_confs":
                    img = self._compute_confidence_gt(preds["flows"], inputs["flows"])
                elif name.startswith("epe"):
                    epe = torch.norm(preds[k] - inputs[k], p=2, dim=2, keepdim=True)
                    img = torch.clamp(epe, 0, self.epe_clip) / self.epe_clip
                    if inputs.get("valids") is not None:
                        img[inputs["valids"] < 0.5] = 0
                else:
                    img = source[k]

                img = img[:1, 0].detach().cpu()
                img = F.interpolate(img, self.image_size)
                img = img[0]

                if "images" in name:
                    img = img.flip([0])  # BGR to RGB
                elif "flows" in name:
                    img = flow_utils.flow_to_rgb(img)

                images[name].append(img)

    def _compute_confidence_gt(
        self, pred_flows: torch.Tensor, target_flows: torch.Tensor
    ) -> torch.Tensor:
        """Compute a confidence score for the flow predictions.

        This score was proposed in https://arxiv.org/abs/2007.09319.

        Parameters
        ----------
        pred_flows : torch.Tensor
            The predicted optical flow.
        target_flows : torch.Tensor
            The groundtruth optical flow.

        Returns
        -------
        torch.Tensor
            The confidence score for each pixel of the input.
        """
        conf_gt = torch.exp(
            -torch.pow(pred_flows - target_flows, 2).sum(dim=2, keepdim=True)
        )
        return conf_gt

    def _compute_max_range(
        self, dataloader_length: int, limit_batches: Union[float, int]
    ) -> int:
        """Find the maximum number of samples that will be drawn from a dataloader.

        Parameters
        ----------
        dataloader_length : int
            Total size of the dataloader.
        limit_batches : Union[float, int]
            A value that may decrease the samples in the dataloader. See --limit_val_batches or --limit_train_batches from
            PyTorch Lightning for more information.

        Returns
        -------
        int
            The maximum number of samples that will be drawn from the dataloader.
        """
        if isinstance(limit_batches, int):
            max_range = limit_batches - 1
        else:
            max_range = int(limit_batches * dataloader_length) - 1
        return max_range

    def _make_image_grid(
        self, dl_images: Dict[str, List[torch.Tensor]]
    ) -> torch.Tensor:
        """Transform a bunch of images into a single one by adding them to a grid.

        Parameters
        ----------
        dl_images : Dict[str, List[torch.Tensor]]
            Lists of images, each identified by a title name.

        Returns
        -------
        torch.Tensor
            A single 3D tensor image 3HW.
        """
        imgs = []
        for img_label, img_list in dl_images.items():
            for j, im in enumerate(img_list):
                if len(im.shape) == 2:
                    im = im[None]

                if im.shape[0] == 1:
                    im = im.repeat(3, 1, 1)

                if j == 0:
                    im = self._add_title(im, img_label)

                imgs.append(im)

        grid = None
        if len(dl_images) > 0:
            grid = make_grid(imgs, len(imgs) // len(dl_images))
        return grid



# functions from timm
from .drop import DropBlock2d, DropPath, drop_block_2d, drop_path
from .helpers import to_ntuple, to_2tuple, to_3tuple, to_4tuple, make_divisible
from .weight_init import trunc_normal_, variance_scaling_, lecun_normal_



from typing import Optional

from loguru import logger
import torch
import torch.nn as nn
import torch.nn.functional as F

from ptlflow.utils.registry import register_model, trainable
from ptlflow.utils.utils import forward_interpolate_batch
from .update import BasicUpdateBlock, GMAUpdateBlock
from .extractor import BasicEncoder
from .matching_encoder import MatchingModel
from .corr import get_corr_block
from .utils import coords_grid, upflow8, compute_grid_indices, compute_weight
from .gma import Attention
from ..base_model.base_model import BaseModel

try:
    import alt_cuda_corr
except:
    alt_cuda_corr = None


class SequenceLoss(nn.Module):
    def __init__(self, gamma: float, max_flow: float):
        super().__init__()
        self.gamma = gamma
        self.max_flow = max_flow

    def forward(self, outputs, inputs):
        """Loss function defined over sequence of flow predictions"""

        flow_preds = outputs["flow_preds"]
        flow_gt = inputs["flows"][:, 0]
        valid = inputs["valids"][:, 0]

        n_predictions = len(flow_preds)
        flow_loss = 0.0

        # exlude invalid pixels and extremely large diplacements
        mag = torch.sum(flow_gt**2, dim=1, keepdim=True).sqrt()
        valid = (valid >= 0.5) & (mag < self.max_flow)

        for i in range(n_predictions):
            i_weight = self.gamma ** (n_predictions - i - 1)
            i_loss = (flow_preds[i] - flow_gt).abs()
            flow_loss += i_weight * (valid * i_loss).mean()

        return flow_loss


class MatchFlow(BaseModel):
    pretrained_checkpoints = {
        "chairs": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/matchflow_gma-chairs-02519b53.ckpt",
        "kitti": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/matchflow_gma-kitti-bc72ce81.ckpt",
        "sintel": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/matchflow_gma-sintel-683422f4.ckpt",
        "things": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/matchflow_gma-things-49295bd8.ckpt",
    }

    def __init__(
        self,
        corr_levels: int = 4,
        corr_radius: int = 4,
        dropout: float = 0.0,
        gamma: float = 0.8,
        max_flow: float = 400,
        iters: int = 32,
        matching_model_path: str = "",
        num_heads: int = 1,
        raft: bool = False,
        use_tile_input: bool = True,
        tile_height: int = 416,
        tile_sigma: float = 0.05,
        position_only: bool = False,
        position_and_content: bool = False,
        alternate_corr: bool = False,
        train_size: Optional[tuple[int, int]] = None,
        **kwargs,
    ) -> None:
        super().__init__(
            output_stride=32, loss_fn=SequenceLoss(gamma, max_flow), **kwargs
        )

        self.corr_levels = corr_levels
        self.corr_radius = corr_radius
        self.dropout = dropout
        self.gamma = gamma
        self.iters = iters
        self.matching_model_path = matching_model_path
        self.num_heads = num_heads
        self.raft = raft
        self.use_tile_input = use_tile_input
        self.tile_height = tile_height
        self.tile_sigma = tile_sigma
        self.position_only = position_only
        self.position_and_content = position_and_content
        self.alternate_corr = alternate_corr
        self.train_size = train_size

        self.hidden_dim = hdim = 128
        self.context_dim = cdim = 128

        # feature network, context network, and update block
        self.fnet = MatchingModel()

        self.cnet = BasicEncoder(
            output_dim=hdim + cdim, norm_fn="batch", dropout=dropout
        )
        if self.raft is False:
            self.update_block = GMAUpdateBlock(
                corr_levels=corr_levels,
                corr_radius=corr_radius,
                num_heads=num_heads,
                hidden_dim=hdim,
            )
            self.att = Attention(
                position_only=position_only,
                position_and_content=position_and_content,
                dim=cdim,
                heads=self.num_heads,
                max_pos_size=160,
                dim_head=cdim,
            )
        else:
            self.update_block = BasicUpdateBlock(
                corr_levels=corr_levels, corr_radius=corr_radius, hidden_dim=hdim
            )

        if self.alternate_corr and alt_cuda_corr is None:
            logger.warning(
                "!!! alt_cuda_corr is not compiled! The slower IterativeCorrBlock will be used instead !!!"
            )

        self.showed_warning = False

    @property
    def train_size(self):
        return self._train_size

    @train_size.setter
    def train_size(self, value):
        if value is not None:
            assert isinstance(value, (tuple, list))
            assert len(value) == 2
            assert isinstance(value[0], int) and isinstance(value[1], int)
        self._train_size = value
        self.fnet = MatchingModel(train_size=self.train_size)

    def freeze_bn(self):
        for m in self.modules():
            if isinstance(m, nn.BatchNorm2d):
                m.eval()

    def initialize_flow(self, img):
        """Flow is represented as difference between two coordinate grids flow = coords1 - coords0"""
        N, C, H, W = img.shape
        coords0 = coords_grid(N, H // 8, W // 8, dtype=img.dtype, device=img.device)
        coords1 = coords_grid(N, H // 8, W // 8, dtype=img.dtype, device=img.device)

        # optical flow computed as difference: flow = coords1 - coords0
        return coords0, coords1

    def upsample_flow(self, flow, mask):
        """Upsample flow field [H/8, W/8, 2] -> [H, W, 2] using convex combination"""
        N, _, H, W = flow.shape
        mask = mask.view(N, 1, 9, 8, 8, H, W)
        mask = torch.softmax(mask, dim=2)

        up_flow = F.unfold(8 * flow, [3, 3], padding=1)
        up_flow = up_flow.view(N, 2, 9, 1, 1, H, W)

        up_flow = torch.sum(mask * up_flow, dim=2)
        up_flow = up_flow.permute(0, 1, 4, 2, 5, 3)
        return up_flow.reshape(N, 2, 8 * H, 8 * W)

    def forward(self, inputs):
        """Estimate optical flow between pair of frames"""
        if self.train_size is not None:
            train_size = self.train_size
        else:
            train_size = self.train_size

        if self.use_tile_input and train_size is None and not self.showed_warning:
            logger.warning(
                "--train_size is not provided and it cannot be loaded from the checkpoint either. Matchflow will run without input tile."
            )
            self.showed_warning = True

        if self.use_tile_input and train_size is not None:
            return self.forward_tile(inputs, train_size)
        else:
            return self.forward_resize(inputs)

    def forward_resize(self, inputs):
        images, image_resizer = self.preprocess_images(
            inputs["images"],
            bgr_add=-0.5,
            bgr_mult=2.0,
            bgr_to_rgb=True,
            resize_mode="interpolation",
            interpolation_mode="bilinear",
            interpolation_align_corners=True,
        )

        flow_prev = None
        if (
            inputs.get("prev_preds") is not None
            and inputs["prev_preds"].get("flow_small") is not None
        ):
            flow_prev = inputs["prev_preds"]["flow_small"]

        flow_predictions, flow_small = self.predict(
            images[:, 0], images[:, 1], flow_prev
        )
        output_flow = flow_predictions[-1]

        if self.training:
            for i, p in enumerate(flow_predictions):
                flow_predictions[i] = self.postprocess_predictions(
                    p, image_resizer, is_flow=True
                )
            outputs = {
                "flows": flow_predictions[-1][:, None],
                "flow_preds": flow_predictions,
            }
        else:
            output_flow = self.postprocess_predictions(
                output_flow, image_resizer, is_flow=True
            )
            outputs = {"flows": output_flow[:, None], "flow_small": flow_small}

        return outputs

    def forward_tile(self, inputs, train_size):
        input_size = inputs["images"].shape[-2:]
        image_size = (max(self.tile_height, input_size[-2]), input_size[-1])
        hws = compute_grid_indices(image_size, train_size)
        device = inputs["images"].device
        weights = compute_weight(
            hws, image_size, train_size, self.tile_sigma, device=device
        )

        images, image_resizer = self.preprocess_images(
            inputs["images"],
            bgr_add=-0.5,
            bgr_mult=2.0,
            bgr_to_rgb=True,
            resize_mode="interpolation",
            target_size=image_size,
            interpolation_mode="bilinear",
            interpolation_align_corners=True,
        )

        image1 = images[:, 0]
        image2 = images[:, 1]

        flows = 0
        flow_count = 0

        for idx, (h, w) in enumerate(hws):
            image1_tile = image1[:, :, h : h + train_size[0], w : w + train_size[1]]
            image2_tile = image2[:, :, h : h + train_size[0], w : w + train_size[1]]

            flow_predictions, _ = self.predict(image1_tile, image2_tile)
            flow_pre = flow_predictions[-1]

            padding = (
                w,
                image_size[1] - w - train_size[1],
                h,
                image_size[0] - h - train_size[0],
                0,
                0,
            )
            flows += F.pad(flow_pre * weights[idx], padding)
            flow_count += F.pad(weights[idx], padding)

        output_flow = flows / flow_count

        output_flow = self.postprocess_predictions(
            output_flow, image_resizer, is_flow=True
        )
        return {"flows": output_flow[:, None]}

    def predict(self, image1, image2, flow_prev=None):
        hdim = self.hidden_dim
        cdim = self.context_dim

        # run the feature network
        fmap1, fmap2 = self.fnet([image1, image2])

        fmap1 = fmap1.float()
        fmap2 = fmap2.float()
        corr_fn = get_corr_block(
            fmap1=fmap1,
            fmap2=fmap2,
            radius=self.corr_radius,
            num_levels=self.corr_levels,
            alternate_corr=self.alternate_corr,
        )

        # run the context network
        cnet = self.cnet(image1)
        net, inp = torch.split(cnet, [hdim, cdim], dim=1)
        net = torch.tanh(net)
        inp = torch.relu(inp)
        # attention, att_c, att_p = self.att(inp)
        if self.raft is False:
            attention = self.att(inp)

        coords0, coords1 = self.initialize_flow(image1)

        if flow_prev is not None:
            forward_flow = forward_interpolate_batch(flow_prev)
            coords1 = coords1 + forward_flow

        flow_predictions = []
        for itr in range(self.iters):
            coords1 = coords1.detach()
            corr = corr_fn(coords1)  # index correlation volume

            flow = coords1 - coords0
            if self.raft is False:
                net, up_mask, delta_flow = self.update_block(
                    net, inp, corr, flow, attention
                )
            else:
                net, up_mask, delta_flow = self.update_block(net, inp, corr, flow)

            # F(t+1) = F(t) + \Delta(t)
            coords1 = coords1 + delta_flow

            # upsample predictions
            if up_mask is None:
                flow_up = upflow8(coords1 - coords0)
            else:
                flow_up = self.upsample_flow(coords1 - coords0, up_mask)

            flow_predictions.append(flow_up)

        return flow_predictions, coords1 - coords0


class MatchFlowRAFT(MatchFlow):
    pretrained_checkpoints = {
        "things": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/matchflow_raft-things-bf560032.ckpt"
    }

    def __init__(
        self,
        corr_levels: int = 4,
        corr_radius: int = 4,
        dropout: float = 0,
        gamma: float = 0.8,
        max_flow: float = 400,
        iters: int = 32,
        matching_model_path: str = "",
        num_heads: int = 1,
        raft: bool = True,
        use_tile_input: bool = True,
        tile_height: int = 416,
        tile_sigma: float = 0.05,
        position_only: bool = False,
        position_and_content: bool = False,
        alternate_corr: bool = False,
        train_size: tuple[int, int] | None = None,
        **kwargs,
    ) -> None:
        super().__init__(
            corr_levels,
            corr_radius,
            dropout,
            gamma,
            max_flow,
            iters,
            matching_model_path,
            num_heads,
            raft,
            use_tile_input,
            tile_height,
            tile_sigma,
            position_only,
            position_and_content,
            alternate_corr,
            train_size,
            **kwargs,
        )


@register_model
@trainable
class matchflow(MatchFlow):
    pass


@register_model
@trainable
class matchflow_raft(MatchFlowRAFT):
    pass



# =============================================================================
# Copyright 2023 Henrique Morimitsu
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Adapted from VAN: https://github.com/Visual-Attention-Network/VAN-Classification/blob/main/models/van.py
# =============================================================================

import math

import torch
import torch.nn as nn

from .local_timm.drop import DropPath
from .local_timm.layer_helpers import to_2tuple
from .local_timm.norm import GroupNorm
from .local_timm.weight_init import trunc_normal_
from .pkconv import PKConv2d


class Mlp(nn.Module):
    def __init__(
        self,
        in_features,
        hidden_features=None,
        out_features=None,
        act_layer=nn.GELU,
        drop=0.0,
        skip_dw=False,
        cache_pkconv_weights=False,
    ):
        super().__init__()
        out_features = out_features or in_features
        hidden_features = hidden_features or in_features
        self.fc1 = PKConv2d(
            in_features, hidden_features, 1, cache_weights=cache_pkconv_weights
        )
        self.dwconv = None
        if not skip_dw:
            self.dwconv = DWConv(
                hidden_features, cache_pkconv_weights=cache_pkconv_weights
            )
        self.act = act_layer()
        self.fc2 = PKConv2d(
            hidden_features, out_features, 1, cache_weights=cache_pkconv_weights
        )
        self.drop = nn.Dropout(drop)

        self.in_hid_factor = float(hidden_features) / in_features
        self.hid_out_factor = float(out_features) / hidden_features

        self.apply(self._init_weights)

    def _init_weights(self, m):
        if isinstance(m, nn.Linear):
            trunc_normal_(m.weight, std=0.02)
            if isinstance(m, nn.Linear) and m.bias is not None:
                nn.init.constant_(m.bias, 0)
        elif isinstance(m, nn.LayerNorm):
            if m.bias is not None:
                nn.init.constant_(m.bias, 0)
            if m.weight is not None:
                nn.init.constant_(m.weight, 1.0)
        elif isinstance(m, nn.Conv2d):
            fan_out = m.kernel_size[0] * m.kernel_size[1] * m.out_channels
            fan_out //= m.groups
            m.weight.data.normal_(0, math.sqrt(2.0 / fan_out))
            if m.bias is not None:
                m.bias.data.zero_()

    def forward(self, x):
        out_ch = int(self.in_hid_factor * x.shape[1])
        x = self.fc1(x, out_ch=out_ch)
        if self.dwconv is not None:
            x = self.dwconv(x, out_ch=out_ch)
        x = self.act(x)
        x = self.drop(x)
        out_ch = int(self.hid_out_factor * x.shape[1])
        x = self.fc2(x, out_ch=out_ch)
        x = self.drop(x)
        return x


class SLKUnitCore(nn.Module):
    def __init__(
        self,
        dim,
        ksize=23,
        cache_pkconv_weights=False,
    ):
        super().__init__()
        self.conv1_branches = nn.ModuleList()
        self.conv1_branches.append(
            PKConv2d(
                dim,
                dim,
                (ksize, 1),
                padding=(ksize // 2, 0),
                groups=dim,
                cache_weights=cache_pkconv_weights,
            )
        )
        self.conv2_branches = nn.ModuleList()
        self.conv2_branches.append(
            PKConv2d(
                dim,
                dim,
                (1, ksize),
                padding=(0, ksize // 2),
                groups=dim,
                cache_weights=cache_pkconv_weights,
            )
        )
        self.conv_out = PKConv2d(dim, dim, 1, cache_weights=cache_pkconv_weights)

    def forward(self, x, out_ch=None):
        y = x
        y = y + self.conv1_branches[0](y, out_ch=out_ch)
        y = y + self.conv2_branches[0](y, out_ch=out_ch)
        y = self.conv_out(y, out_ch=out_ch)
        y = y + x
        return y


class SLKUnit(nn.Module):
    def __init__(
        self,
        dim,
        cache_pkconv_weights=False,
    ):
        super().__init__()

        self.proj_1 = PKConv2d(dim, dim, 1, cache_weights=cache_pkconv_weights)
        self.activation = nn.GELU()
        self.spatial_gating_unit = SLKUnitCore(
            dim,
            cache_pkconv_weights=cache_pkconv_weights,
        )
        self.proj_2 = PKConv2d(dim, dim, 1, cache_weights=cache_pkconv_weights)

    def forward(self, x):
        out_ch = x.shape[1]
        shorcut = x.clone()
        x = self.proj_1(x, out_ch=out_ch)
        x = self.activation(x)
        x = self.spatial_gating_unit(x, out_ch=out_ch)
        x = self.proj_2(x, out_ch=out_ch)
        x = x + shorcut
        return x


class SLK(nn.Module):
    def __init__(
        self,
        dim,
        mlp_ratio=4.0,
        drop=0.0,
        drop_path=0.0,
        act_layer=nn.GELU,
        norm_layer=GroupNorm,
        cache_pkconv_weights=False,
    ):
        super().__init__()
        self.norm1 = norm_layer(num_channels=dim)
        self.attn = SLKUnit(
            dim,
            cache_pkconv_weights=cache_pkconv_weights,
        )
        self.drop_path = DropPath(drop_path) if drop_path > 0.0 else nn.Identity()

        self.norm2 = norm_layer(num_channels=dim)
        mlp_hidden_dim = int(dim * mlp_ratio)
        self.mlp = Mlp(
            in_features=dim,
            hidden_features=mlp_hidden_dim,
            act_layer=act_layer,
            drop=drop,
            cache_pkconv_weights=cache_pkconv_weights,
        )
        layer_scale_init_value = 1e-2
        self.layer_scale_1 = nn.Parameter(
            layer_scale_init_value * torch.ones((dim)), requires_grad=True
        )
        self.layer_scale_2 = nn.Parameter(
            layer_scale_init_value * torch.ones((dim)), requires_grad=True
        )

        self.apply(self._init_weights)

    def _init_weights(self, m):
        if isinstance(m, nn.Linear):
            trunc_normal_(m.weight, std=0.02)
            if isinstance(m, nn.Linear) and m.bias is not None:
                nn.init.constant_(m.bias, 0)
        elif isinstance(m, nn.LayerNorm):
            if m.bias is not None:
                nn.init.constant_(m.bias, 0)
            if m.weight is not None:
                nn.init.constant_(m.weight, 1.0)
        elif isinstance(m, nn.Conv2d):
            fan_out = m.kernel_size[0] * m.kernel_size[1] * m.out_channels
            fan_out //= m.groups
            m.weight.data.normal_(0, math.sqrt(2.0 / fan_out))
            if m.bias is not None:
                m.bias.data.zero_()

    def forward(self, x, out_ch=None):
        x = x + self.drop_path(
            self.layer_scale_1[: x.shape[1]].unsqueeze(-1).unsqueeze(-1)
            * self.attn(self.norm1(x))
        )
        x = x + self.drop_path(
            self.layer_scale_2[: x.shape[1]].unsqueeze(-1).unsqueeze(-1)
            * self.mlp(self.norm2(x))
        )
        return x


class LayerTransition(nn.Module):
    """Image to Patch Embedding"""

    def __init__(
        self,
        patch_size=3,
        stride=2,
        in_chans=64,
        embed_dim=64,
        norm_layer=GroupNorm,
        cache_pkconv_weights=False,
    ):
        super().__init__()
        patch_size = to_2tuple(patch_size)
        self.proj = PKConv2d(
            in_chans,
            embed_dim,
            kernel_size=patch_size,
            stride=stride,
            padding=(patch_size[0] // 2, patch_size[1] // 2),
            cache_weights=cache_pkconv_weights,
        )
        self.norm = norm_layer(num_channels=embed_dim)

        self.apply(self._init_weights)

    def _init_weights(self, m):
        if isinstance(m, nn.Linear):
            trunc_normal_(m.weight, std=0.02)
            if isinstance(m, nn.Linear) and m.bias is not None:
                nn.init.constant_(m.bias, 0)
        elif isinstance(m, nn.LayerNorm):
            if m.bias is not None:
                nn.init.constant_(m.bias, 0)
            if m.weight is not None:
                nn.init.constant_(m.weight, 1.0)
        elif isinstance(m, nn.Conv2d):
            fan_out = m.kernel_size[0] * m.kernel_size[1] * m.out_channels
            fan_out //= m.groups
            m.weight.data.normal_(0, math.sqrt(2.0 / fan_out))
            if m.bias is not None:
                m.bias.data.zero_()

    def forward(self, x, out_ch=None):
        x = self.proj(x, out_ch=out_ch)
        x = self.norm(x)
        return x


class DWConv(nn.Module):
    def __init__(
        self,
        dim=768,
        cache_pkconv_weights=False,
    ):
        super(DWConv, self).__init__()
        self.dwconv = PKConv2d(
            dim, dim, 3, 1, 1, bias=True, groups=dim, cache_weights=cache_pkconv_weights
        )

    def forward(self, x, out_ch=None):
        x = self.dwconv(x, out_ch=out_ch)
        return x


class PKConvSLK(nn.Module):
    def __init__(
        self,
        in_chs,
        out_chs,
        mlp_ratio=4.0,
        drop=0.0,
        drop_path=0.0,
        act_layer=nn.GELU,
        norm_layer=GroupNorm,
        stride=1,
        depth=2,
        cache_pkconv_weights=False,
    ):
        super(PKConvSLK, self).__init__()
        self.down = None
        if stride > 1 or in_chs != out_chs:
            patch_size = 1
            if stride > 1:
                patch_size = 3
            self.down = LayerTransition(
                patch_size=patch_size,
                stride=stride,
                in_chans=in_chs,
                embed_dim=out_chs,
                norm_layer=norm_layer,
                cache_pkconv_weights=cache_pkconv_weights,
            )

        self.blocks = nn.ModuleList()
        for _ in range(depth):
            self.blocks.append(
                SLK(
                    dim=out_chs,
                    mlp_ratio=mlp_ratio,
                    drop=drop,
                    drop_path=drop_path,
                    act_layer=act_layer,
                    norm_layer=norm_layer,
                    cache_pkconv_weights=cache_pkconv_weights,
                )
            )
        self.norm = norm_layer(num_channels=out_chs)

    def forward(self, x, out_ch=None):
        if self.down is not None:
            x = self.down(x, out_ch=out_ch)
        for blk in self.blocks:
            x = blk(x, out_ch=out_ch)
        x = self.norm(x)
        return x



from einops import rearrange

import torch
import torch.nn as nn
from torch.nn import init

from ptlflow.utils.registry import register_model, trainable
from .submodules import *
from .flownet_base import FlowNetBase


class FlowNetC(FlowNetBase):
    pretrained_checkpoints = {
        "things": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/flownetc-things-cc8ac7fd.ckpt"
    }

    def __init__(
        self,
        div_flow: float = 20.0,
        input_channels: int = 6,
        batch_norm: bool = False,
        loss_start_scale: int = 4,
        loss_num_scales: int = 5,
        loss_base_weight: float = 0.32,
        loss_norm: str = "L2",
        **kwargs,
    ):
        super(FlowNetC, self).__init__(
            div_flow=div_flow,
            input_channels=input_channels,
            batch_norm=batch_norm,
            loss_start_scale=loss_start_scale,
            loss_num_scales=loss_num_scales,
            loss_base_weight=loss_base_weight,
            loss_norm=loss_norm,
            **kwargs,
        )

        self.rgb_max = 1

        self.conv1 = conv(self.batch_norm, 3, 64, kernel_size=7, stride=2)
        self.conv2 = conv(self.batch_norm, 64, 128, kernel_size=5, stride=2)
        self.conv3 = conv(self.batch_norm, 128, 256, kernel_size=5, stride=2)
        self.conv_redir = conv(self.batch_norm, 256, 32, kernel_size=1, stride=1)

        self.corr = correlate  # Correlation(pad_size=20, kernel_size=1, max_displacement=20, stride1=1, stride2=2, corr_multiply=1)

        self.corr_activation = nn.LeakyReLU(0.1, inplace=True)
        self.conv3_1 = conv(self.batch_norm, 473, 256)
        self.conv4 = conv(self.batch_norm, 256, 512, stride=2)
        self.conv4_1 = conv(self.batch_norm, 512, 512)
        self.conv5 = conv(self.batch_norm, 512, 512, stride=2)
        self.conv5_1 = conv(self.batch_norm, 512, 512)
        self.conv6 = conv(self.batch_norm, 512, 1024, stride=2)
        self.conv6_1 = conv(self.batch_norm, 1024, 1024)

        self.deconv5 = deconv(1024, 512)
        self.deconv4 = deconv(1026, 256)
        self.deconv3 = deconv(770, 128)
        self.deconv2 = deconv(386, 64)

        self.predict_flow6 = predict_flow(1024)
        self.predict_flow5 = predict_flow(1026)
        self.predict_flow4 = predict_flow(770)
        self.predict_flow3 = predict_flow(386)
        self.predict_flow2 = predict_flow(194)

        self.upsampled_flow6_to_5 = nn.ConvTranspose2d(2, 2, 4, 2, 1, bias=True)
        self.upsampled_flow5_to_4 = nn.ConvTranspose2d(2, 2, 4, 2, 1, bias=True)
        self.upsampled_flow4_to_3 = nn.ConvTranspose2d(2, 2, 4, 2, 1, bias=True)
        self.upsampled_flow3_to_2 = nn.ConvTranspose2d(2, 2, 4, 2, 1, bias=True)

        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                if m.bias is not None:
                    init.uniform_(m.bias)
                init.xavier_uniform_(m.weight)

            if isinstance(m, nn.ConvTranspose2d):
                if m.bias is not None:
                    init.uniform_(m.bias)
                init.xavier_uniform_(m.weight)
                # init_deconv_bilinear(m.weight)
        self.upsample1 = nn.Upsample(
            scale_factor=4, mode="bilinear", align_corners=False
        )

    def forward(self, inputs, skip_preprocess=False):
        images = inputs["images"]
        if skip_preprocess:
            image_resizer = None
        else:
            bgr_mean = rearrange(images, "b n c h w -> b n c (h w)").mean(-1)
            bgr_mean = bgr_mean[..., None, None]
            images, image_resizer = self.preprocess_images(
                images,
                bgr_add=-bgr_mean,
                bgr_mult=1.0,
                bgr_to_rgb=True,
                resize_mode="interpolation",
                interpolation_mode="bilinear",
                interpolation_align_corners=True,
            )
        x1 = images[:, 0]
        x2 = images[:, 1]

        # FlownetC top input stream
        out_conv1a = self.conv1(x1)
        out_conv2a = self.conv2(out_conv1a)
        out_conv3a = self.conv3(out_conv2a)

        # FlownetC bottom input stream
        out_conv1b = self.conv1(x2)

        out_conv2b = self.conv2(out_conv1b)
        out_conv3b = self.conv3(out_conv2b)

        # Merge streams
        out_corr = self.corr(out_conv3a, out_conv3b)  # False
        out_corr = self.corr_activation(out_corr)

        # Redirect top input stream and concatenate
        out_conv_redir = self.conv_redir(out_conv3a)

        in_conv3_1 = torch.cat((out_conv_redir, out_corr), 1)

        # Merged conv layers
        out_conv3_1 = self.conv3_1(in_conv3_1)

        out_conv4 = self.conv4_1(self.conv4(out_conv3_1))

        out_conv5 = self.conv5_1(self.conv5(out_conv4))
        out_conv6 = self.conv6_1(self.conv6(out_conv5))

        flow6 = self.predict_flow6(out_conv6)
        flow6_up = self.upsampled_flow6_to_5(flow6)
        out_deconv5 = self.deconv5(out_conv6)

        concat5 = torch.cat((out_conv5, out_deconv5, flow6_up), 1)

        flow5 = self.predict_flow5(concat5)
        flow5_up = self.upsampled_flow5_to_4(flow5)
        out_deconv4 = self.deconv4(concat5)
        concat4 = torch.cat((out_conv4, out_deconv4, flow5_up), 1)

        flow4 = self.predict_flow4(concat4)
        flow4_up = self.upsampled_flow4_to_3(flow4)
        out_deconv3 = self.deconv3(concat4)
        concat3 = torch.cat((out_conv3_1, out_deconv3, flow4_up), 1)

        flow3 = self.predict_flow3(concat3)
        flow3_up = self.upsampled_flow3_to_2(flow3)
        out_deconv2 = self.deconv2(concat3)
        concat2 = torch.cat((out_conv2a, out_deconv2, flow3_up), 1)

        flow2 = self.predict_flow2(concat2)

        out_flow = self.div_flow * self.upsample1(flow2)
        if image_resizer is not None:
            out_flow = self.postprocess_predictions(
                out_flow, image_resizer, is_flow=True
            )

        outputs = {}

        if self.training:
            outputs["flow_preds"] = [flow2, flow3, flow4, flow5, flow6]
            outputs["flows"] = out_flow[:, None]
        else:
            outputs["flows"] = out_flow[:, None]

        return outputs


@register_model
@trainable
class flownetc(FlowNetC):
    pass



import torch
import torch.nn as nn
import torch.nn.functional as F

from einops import rearrange

from ptlflow.utils.utils import forward_interpolate_batch
from .utils import coords_grid, bilinear_sampler
from .attention import (
    MultiHeadAttention,
    LinearPositionEmbeddingSine,
    ExpPositionEmbeddingSine,
)

from timm.layers import DropPath

from .gru import BasicUpdateBlock, GMAUpdateBlock
from .gma import Attention


def initialize_flow(img):
    """Flow is represented as difference between two means flow = mean1 - mean0"""
    N, C, H, W = img.shape
    mean = coords_grid(N, H, W, dtype=img.dtype, device=img.device)
    mean_init = coords_grid(N, H, W, dtype=img.dtype, device=img.device)

    # optical flow computed as difference: flow = mean1 - mean0
    return mean, mean_init


class CrossAttentionLayer(nn.Module):
    # def __init__(self, dim, cfg, num_heads=8, attn_drop=0., proj_drop=0., drop_path=0., dropout=0.):
    def __init__(
        self,
        qk_dim,
        v_dim,
        query_token_dim,
        tgt_token_dim,
        add_flow_token=True,
        num_heads=8,
        attn_drop=0.0,
        proj_drop=0.0,
        drop_path=0.0,
        dropout=0.0,
        pe="linear",
    ):
        super(CrossAttentionLayer, self).__init__()

        head_dim = qk_dim // num_heads
        self.scale = head_dim**-0.5
        self.query_token_dim = query_token_dim
        self.pe = pe

        self.norm1 = nn.LayerNorm(query_token_dim)
        self.norm2 = nn.LayerNorm(query_token_dim)
        self.multi_head_attn = MultiHeadAttention(qk_dim, num_heads)
        self.q, self.k, self.v = (
            nn.Linear(query_token_dim, qk_dim, bias=True),
            nn.Linear(tgt_token_dim, qk_dim, bias=True),
            nn.Linear(tgt_token_dim, v_dim, bias=True),
        )

        self.proj = nn.Linear(v_dim * 2, query_token_dim)
        self.proj_drop = nn.Dropout(proj_drop)
        self.drop_path = DropPath(drop_path) if drop_path > 0.0 else nn.Identity()

        self.ffn = nn.Sequential(
            nn.Linear(query_token_dim, query_token_dim),
            nn.GELU(),
            nn.Dropout(dropout),
            nn.Linear(query_token_dim, query_token_dim),
            nn.Dropout(dropout),
        )
        self.add_flow_token = add_flow_token
        self.dim = qk_dim

    def forward(self, query, key, value, memory, query_coord, patch_size, size_h3w3):
        """
        query_coord [B, 2, H1, W1]
        """
        B, _, H1, W1 = query_coord.shape

        if key is None and value is None:
            key = self.k(memory)
            value = self.v(memory)

        # [B, 2, H1, W1] -> [BH1W1, 1, 2]
        query_coord = query_coord.contiguous()
        query_coord = (
            query_coord.view(B, 2, -1)
            .permute(0, 2, 1)[:, :, None, :]
            .contiguous()
            .view(B * H1 * W1, 1, 2)
        )
        if self.pe == "linear":
            query_coord_enc = LinearPositionEmbeddingSine(query_coord, dim=self.dim)
        elif self.pe == "exp":
            query_coord_enc = ExpPositionEmbeddingSine(query_coord, dim=self.dim)

        short_cut = query
        query = self.norm1(query)

        if self.add_flow_token:
            q = self.q(query + query_coord_enc)
        else:
            q = self.q(query_coord_enc)
        k, v = key, value

        x = self.multi_head_attn(q, k, v)

        x = self.proj(torch.cat([x, short_cut], dim=2))
        x = short_cut + self.proj_drop(x)

        x = x + self.drop_path(self.ffn(self.norm2(x)))

        return x, k, v


class MemoryDecoderLayer(nn.Module):
    def __init__(
        self, patch_size, query_latent_dim, cost_latent_dim, add_flow_token, dropout
    ):
        super(MemoryDecoderLayer, self).__init__()
        self.patch_size = patch_size  # for converting coords into H2', W2' space
        self.query_latent_dim = query_latent_dim

        query_token_dim, tgt_token_dim = query_latent_dim, cost_latent_dim
        qk_dim, v_dim = query_token_dim, query_token_dim
        self.cross_attend = CrossAttentionLayer(
            qk_dim,
            v_dim,
            query_token_dim,
            tgt_token_dim,
            add_flow_token=add_flow_token,
            dropout=dropout,
        )

    def forward(self, query, key, value, memory, coords1, size, size_h3w3):
        """
        x:      [B*H1*W1, 1, C]
        memory: [B*H1*W1, H2'*W2', C]
        coords1 [B, 2, H2, W2]
        size: B, C, H1, W1
        1. Note that here coords0 and coords1 are in H2, W2 space.
           Should first convert it into H2', W2' space.
        2. We assume the upper-left point to be [0, 0], instead of letting center of upper-left patch to be [0, 0]
        """
        x_global, k, v = self.cross_attend(
            query, key, value, memory, coords1, self.patch_size, size_h3w3
        )
        B, C, H1, W1 = size
        C = self.query_latent_dim
        x_global = x_global.view(B, H1, W1, C).permute(0, 3, 1, 2)
        return x_global, k, v


class ReverseCostExtractor(nn.Module):
    def __init__(self, cfg):
        super(ReverseCostExtractor, self).__init__()
        self.cfg = cfg

    def forward(self, cost_maps, coords0, coords1):
        """
        cost_maps   -   B*H1*W1, cost_heads_num, H2, W2
        coords      -   B, 2, H1, W1
        """
        BH1W1, heads, H2, W2 = cost_maps.shape
        B, _, H1, W1 = coords1.shape

        assert (H1 == H2) and (W1 == W2)
        assert BH1W1 == B * H1 * W1

        cost_maps = cost_maps.reshape(B, H1 * W1 * heads, H2, W2)
        coords = coords1.permute(0, 2, 3, 1)
        corr = bilinear_sampler(cost_maps, coords)  # [B, H1*W1*heads, H2, W2]
        corr = rearrange(
            corr,
            "b (h1 w1 heads) h2 w2 -> (b h2 w2) heads h1 w1",
            b=B,
            heads=heads,
            h1=H1,
            w1=W1,
            h2=H2,
            w2=W2,
        )

        r = 4
        dx = torch.linspace(-r, r, 2 * r + 1)
        dy = torch.linspace(-r, r, 2 * r + 1)
        delta = torch.stack(torch.meshgrid(dy, dx, indexing="ij"), axis=-1).to(
            coords0.device
        )
        centroid = coords0.permute(0, 2, 3, 1).reshape(BH1W1, 1, 1, 2)
        delta = delta.view(1, 2 * r + 1, 2 * r + 1, 2)
        coords = centroid + delta
        corr = bilinear_sampler(corr, coords)
        corr = corr.view(B, H1, W1, -1).permute(0, 3, 1, 2)
        return corr


class MemoryDecoder(nn.Module):
    def __init__(
        self,
        query_latent_dim,
        cost_heads_num,
        decoder_depth,
        gma,
        only_global,
        patch_size,
        cost_latent_dim,
        add_flow_token,
        dropout,
    ):
        super(MemoryDecoder, self).__init__()
        self.gma = gma
        self.only_global = only_global

        dim = self.dim = query_latent_dim

        self.flow_token_encoder = nn.Sequential(
            nn.Conv2d(81 * cost_heads_num, dim, 1, 1),
            nn.GELU(),
            nn.Conv2d(dim, dim, 1, 1),
        )
        self.proj = nn.Conv2d(256, 256, 1)
        self.depth = decoder_depth
        self.decoder_layer = MemoryDecoderLayer(
            patch_size=patch_size,
            query_latent_dim=query_latent_dim,
            cost_latent_dim=cost_latent_dim,
            add_flow_token=add_flow_token,
            dropout=dropout,
        )

        if gma:
            self.update_block = GMAUpdateBlock(
                only_global=only_global,
                query_latent_dim=query_latent_dim,
                hidden_dim=128,
            )
            self.att = Attention(dim=128, heads=1, max_pos_size=160, dim_head=128)
        else:
            self.update_block = BasicUpdateBlock(
                only_global=only_global,
                query_latent_dim=query_latent_dim,
                hidden_dim=128,
            )

    def upsample_flow(self, flow, mask):
        """Upsample flow field [H/8, W/8, 2] -> [H, W, 2] using convex combination"""
        N, _, H, W = flow.shape
        mask = mask.view(N, 1, 9, 8, 8, H, W)
        mask = torch.softmax(mask, dim=2)

        up_flow = F.unfold(8 * flow, [3, 3], padding=1)
        up_flow = up_flow.view(N, 2, 9, 1, 1, H, W)

        up_flow = torch.sum(mask * up_flow, dim=2)
        up_flow = up_flow.permute(0, 1, 4, 2, 5, 3)
        return up_flow.reshape(N, 2, 8 * H, 8 * W)

    def encode_flow_token(self, cost_maps, coords):
        """
        cost_maps   -   B*H1*W1, cost_heads_num, H2, W2
        coords      -   B, 2, H1, W1
        """
        coords = coords.permute(0, 2, 3, 1)
        batch, h1, w1, _ = coords.shape

        r = 4
        dx = torch.linspace(-r, r, 2 * r + 1, dtype=coords.dtype, device=coords.device)
        dy = torch.linspace(-r, r, 2 * r + 1, dtype=coords.dtype, device=coords.device)
        delta = torch.stack(torch.meshgrid(dy, dx, indexing="ij"), axis=-1)

        centroid = coords.reshape(batch * h1 * w1, 1, 1, 2)
        delta = delta.view(1, 2 * r + 1, 2 * r + 1, 2)
        coords = centroid + delta
        corr = bilinear_sampler(cost_maps, coords)
        corr = corr.view(batch, h1, w1, -1).permute(0, 3, 1, 2)
        return corr

    def forward(self, cost_memory, context, data={}, prev_flow=None):
        """
        memory: [B*H1*W1, H2'*W2', C]
        context: [B, D, H1, W1]
        """
        cost_maps = data["cost_maps"]
        coords0, coords1 = initialize_flow(context)

        if prev_flow is not None:
            forward_flow = forward_interpolate_batch(prev_flow)
            coords1 = coords1 + forward_flow

        # flow = coords1

        flow_predictions = []

        context = self.proj(context)
        net, inp = torch.split(context, [128, 128], dim=1)
        net = torch.tanh(net)
        inp = torch.relu(inp)
        if self.gma:
            attention = self.att(inp)

        size = net.shape
        key, value = None, None

        for idx in range(self.depth):
            coords1 = coords1.detach()

            cost_forward = self.encode_flow_token(cost_maps, coords1)
            # cost_backward = self.reverse_cost_extractor(cost_maps, coords0, coords1)

            query = self.flow_token_encoder(cost_forward)
            query = (
                query.permute(0, 2, 3, 1)
                .contiguous()
                .view(size[0] * size[2] * size[3], 1, self.dim)
            )
            cost_global, key, value = self.decoder_layer(
                query, key, value, cost_memory, coords1, size, data["H3W3"]
            )
            if self.only_global:
                corr = cost_global
            else:
                corr = torch.cat([cost_global, cost_forward], dim=1)

            flow = coords1 - coords0

            if self.gma:
                net, up_mask, delta_flow = self.update_block(
                    net, inp, corr, flow, attention
                )
            else:
                net, up_mask, delta_flow = self.update_block(net, inp, corr, flow)

            # flow = delta_flow
            coords1 = coords1 + delta_flow
            flow_up = self.upsample_flow(coords1 - coords0, up_mask)
            flow_predictions.append(flow_up)

        return flow_predictions, coords1 - coords0



from typing import Sequence

import torch
import torch.nn as nn
import torch.nn.functional as F

from ptlflow.utils.registry import register_model, trainable
from .backbone import CNNEncoder
from .transformer import FeatureTransformer
from .matching import (
    global_correlation_softmax,
    local_correlation_softmax,
    local_correlation_with_flow,
)
from .attention import SelfAttnPropagation
from .geometry import flow_warp
from .reg_refine import BasicUpdateBlock
from .utils import (
    feature_add_position,
    upsample_flow_with_mask,
)
from ..base_model.base_model import BaseModel


class SequenceLoss(nn.Module):
    def __init__(self, gamma: float, max_flow: float):
        super().__init__()
        self.gamma = gamma
        self.max_flow = max_flow

    def forward(self, outputs, inputs):
        """Loss function defined over sequence of flow predictions"""
        flow_preds = outputs["flow_preds"]
        flow_gt = inputs["flows"][:, 0]
        valid = inputs["valids"][:, 0]

        n_predictions = len(flow_preds)
        flow_loss = 0.0

        # exlude invalid pixels and extremely large diplacements
        mag = torch.sum(flow_gt**2, dim=1, keepdim=True).sqrt()
        valid = (valid >= 0.5) & (mag < self.max_flow)

        for i in range(n_predictions):
            i_weight = self.gamma ** (n_predictions - i - 1)
            i_loss = (flow_preds[i] - flow_gt).abs()
            flow_loss += i_weight * (valid * i_loss).mean()

        return flow_loss


class UniMatch(BaseModel):
    pretrained_checkpoints = {
        "mix": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/unimatch-mixdata-9d7c1e4d.ckpt",
        "things": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/unimatch-things-2433864a.ckpt",
    }

    def __init__(
        self,
        gamma: float = 0.9,
        max_flow: float = 400.0,
        corr_radius: int = 4,
        feature_channels: int = 128,
        num_scales: int = 1,
        upsample_factor: int = 8,
        reg_refine: bool = False,
        num_transformer_layers: int = 6,
        num_head: int = 1,
        ffn_dim_expansion: int = 4,
        pred_bidir_flow: bool = False,
        num_reg_refine: int = 1,
        attn_type: str = "swin",
        attn_splits_list: Sequence[int] = (2,),
        corr_radius_list: Sequence[int] = (-1,),
        prop_radius_list: Sequence[int] = (-1,),
        **kwargs,
    ) -> None:
        super().__init__(
            output_stride=32, loss_fn=SequenceLoss(gamma, max_flow), **kwargs
        )

        self.gamma = gamma
        self.max_flow = max_flow
        self.corr_radius = corr_radius
        self.feature_channels = feature_channels
        self.num_scales = num_scales
        self.upsample_factor = upsample_factor
        self.reg_refine = reg_refine
        self.num_transformer_layers = num_transformer_layers
        self.num_head = num_head
        self.ffn_dim_expansion = ffn_dim_expansion
        self.pred_bidir_flow = pred_bidir_flow
        self.num_reg_refine = num_reg_refine
        self.attn_type = attn_type
        self.attn_splits_list = attn_splits_list
        self.corr_radius_list = corr_radius_list
        self.prop_radius_list = prop_radius_list

        # CNN
        self.backbone = CNNEncoder(
            output_dim=self.feature_channels,
            num_output_scales=self.num_scales,
        )

        # Transformer
        self.transformer = FeatureTransformer(
            num_layers=self.num_transformer_layers,
            d_model=self.feature_channels,
            nhead=self.num_head,
            ffn_dim_expansion=self.ffn_dim_expansion,
        )

        # propagation with self-attn
        self.feature_flow_attn = SelfAttnPropagation(in_channels=self.feature_channels)

        if not self.reg_refine:
            # convex upsampling simiar to RAFT
            # concat feature0 and low res flow as input
            self.upsampler = nn.Sequential(
                nn.Conv2d(2 + self.feature_channels, 256, 3, 1, 1),
                nn.ReLU(inplace=True),
                nn.Conv2d(256, self.upsample_factor**2 * 9, 1, 1, 0),
            )
            # thus far, all the learnable parameters are task-agnostic

        if self.reg_refine:
            # optional task-specific local regression refinement
            self.refine_proj = nn.Conv2d(128, 256, 1)
            self.refine = BasicUpdateBlock(
                corr_channels=(2 * 4 + 1) ** 2,
                downsample_factor=self.upsample_factor,
                flow_dim=2,
                bilinear_up=False,
            )

    def extract_feature(self, img0, img1):
        concat = torch.cat((img0, img1), dim=0)  # [2B, C, H, W]
        features = self.backbone(
            concat
        )  # list of [2B, C, H, W], resolution from high to low

        # reverse: resolution from low to high
        features = features[::-1]

        feature0, feature1 = [], []

        for i in range(len(features)):
            feature = features[i]
            chunks = torch.chunk(feature, 2, 0)  # tuple
            feature0.append(chunks[0])
            feature1.append(chunks[1])

        return feature0, feature1

    def upsample_flow(self, flow, feature, bilinear=False, upsample_factor=8):
        if bilinear:
            multiplier = upsample_factor
            up_flow = (
                F.interpolate(
                    flow,
                    scale_factor=upsample_factor,
                    mode="bilinear",
                    align_corners=True,
                )
                * multiplier
            )
        else:
            concat = torch.cat((flow, feature), dim=1)
            mask = self.upsampler(concat)
            up_flow = upsample_flow_with_mask(
                flow, mask, upsample_factor=self.upsample_factor
            )

        return up_flow

    def forward(self, inputs):
        """Estimate optical flow between pair of frames"""
        images, image_resizer = self.preprocess_images(
            inputs["images"],
            bgr_add=[-0.406, -0.456, -0.485],
            bgr_mult=[1 / 0.225, 1 / 0.224, 1 / 0.229],
            bgr_to_rgb=True,
            resize_mode="pad",
            pad_mode="replicate",
            pad_two_side=True,
        )

        img0 = images[:, 0]
        img1 = images[:, 1]

        results_dict = {}
        flow_preds = []

        # list of features, resolution low to high
        feature0_list, feature1_list = self.extract_feature(
            img0, img1
        )  # list of features

        flow = None

        assert (
            len(self.attn_splits_list)
            == len(self.corr_radius_list)
            == len(self.prop_radius_list)
            == self.num_scales
        )

        for scale_idx in range(self.num_scales):
            feature0, feature1 = feature0_list[scale_idx], feature1_list[scale_idx]

            if self.pred_bidir_flow and scale_idx > 0:
                # predicting bidirectional flow with refinement
                feature0, feature1 = torch.cat((feature0, feature1), dim=0), torch.cat(
                    (feature1, feature0), dim=0
                )

            feature0_ori, feature1_ori = feature0, feature1

            upsample_factor = self.upsample_factor * (
                2 ** (self.num_scales - 1 - scale_idx)
            )

            if scale_idx > 0:
                flow = (
                    F.interpolate(
                        flow, scale_factor=2, mode="bilinear", align_corners=True
                    )
                    * 2
                )

            if flow is not None:
                flow = flow.detach()
                feature1 = flow_warp(feature1, flow)  # [B, C, H, W]

            attn_splits = self.attn_splits_list[scale_idx]
            corr_radius = self.corr_radius_list[scale_idx]
            prop_radius = self.prop_radius_list[scale_idx]

            # add position to features
            feature0, feature1 = feature_add_position(
                feature0, feature1, attn_splits, self.feature_channels
            )

            # Transformer
            feature0, feature1 = self.transformer(
                feature0,
                feature1,
                attn_type=self.attn_type,
                attn_num_splits=attn_splits,
            )

            # correlation and softmax
            if corr_radius == -1:  # global matching
                flow_pred = global_correlation_softmax(
                    feature0, feature1, self.pred_bidir_flow
                )[0]
            else:  # local matching
                flow_pred = local_correlation_softmax(feature0, feature1, corr_radius)[
                    0
                ]

            # flow or residual flow
            flow = flow + flow_pred if flow is not None else flow_pred

            # upsample to the original resolution for supervison at training time only
            if self.training:
                flow_bilinear = self.upsample_flow(
                    flow, None, bilinear=True, upsample_factor=upsample_factor
                )
                flow_bilinear = self.postprocess_predictions(
                    flow_bilinear, image_resizer, is_flow=True
                )
                flow_preds.append(flow_bilinear)

            # flow propagation with self-attn
            if self.pred_bidir_flow and scale_idx == 0:
                feature0 = torch.cat(
                    (feature0, feature1), dim=0
                )  # [2*B, C, H, W] for propagation

            flow = self.feature_flow_attn(
                feature0,
                flow.detach(),
                local_window_attn=prop_radius > 0,
                local_window_radius=prop_radius,
            )

            # bilinear exclude the last one
            if self.training and scale_idx < self.num_scales - 1:
                flow_up = self.upsample_flow(
                    flow, feature0, bilinear=True, upsample_factor=upsample_factor
                )
                flow_up = self.postprocess_predictions(
                    flow_up, image_resizer, is_flow=True
                )
                flow_preds.append(flow_up)

            if scale_idx == self.num_scales - 1:
                if not self.reg_refine:
                    # upsample to the original image resolution
                    flow_up = self.upsample_flow(flow, feature0)
                    flow_up = self.postprocess_predictions(
                        flow_up, image_resizer, is_flow=True
                    )
                    flow_preds.append(flow_up)
                else:
                    # task-specific local regression refinement
                    # supervise current flow
                    if self.training:
                        flow_up = self.upsample_flow(
                            flow,
                            feature0,
                            bilinear=True,
                            upsample_factor=upsample_factor,
                        )
                        flow_up = self.postprocess_predictions(
                            flow_up, image_resizer, is_flow=True
                        )
                        flow_preds.append(flow_up)

                    assert self.num_reg_refine > 0
                    for refine_iter_idx in range(self.num_reg_refine):
                        flow = flow.detach()
                        correlation = local_correlation_with_flow(
                            feature0_ori,
                            feature1_ori,
                            flow=flow,
                            local_radius=4,
                        )  # [B, (2R+1)^2, H, W]

                        proj = self.refine_proj(feature0)

                        net, inp = torch.chunk(proj, chunks=2, dim=1)

                        net = torch.tanh(net)
                        inp = torch.relu(inp)

                        net, up_mask, residual_flow = self.refine(
                            net,
                            inp,
                            correlation,
                            flow.clone(),
                        )
                        flow = flow + residual_flow

                        if self.training or refine_iter_idx == self.num_reg_refine - 1:
                            flow_up = upsample_flow_with_mask(
                                flow, up_mask, upsample_factor=self.upsample_factor
                            )

                            flow_up = self.postprocess_predictions(
                                flow_up, image_resizer, is_flow=True
                            )
                            flow_preds.append(flow_up)

        results_dict.update({"flow_preds": flow_preds})

        if self.training:
            outputs = {"flows": flow_up[:, None], "flow_preds": flow_preds}
        else:
            outputs = {"flows": flow_up[:, None], "flow_small": flow}

        return outputs


class UniMatchScale2(UniMatch):
    pretrained_checkpoints = {
        "mix": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/unimatch_scale2-mixdata-b514dde2.ckpt",
        "things": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/unimatch_scale2-things-e75ae2f7.ckpt",
        "sintel": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/unimatch_scale2-sintel-f43b76ab.ckpt",
    }

    def __init__(
        self,
        gamma: float = 0.9,
        max_flow: float = 400,
        corr_radius: int = 4,
        feature_channels: int = 128,
        num_scales: int = 2,
        upsample_factor: int = 4,
        reg_refine: bool = False,
        num_transformer_layers: int = 6,
        num_head: int = 1,
        ffn_dim_expansion: int = 4,
        pred_bidir_flow: bool = False,
        num_reg_refine: int = 1,
        attn_type: str = "swin",
        attn_splits_list: Sequence[int] = (2, 8),
        corr_radius_list: Sequence[int] = (-1, 4),
        prop_radius_list: Sequence[int] = (-1, 1),
        **kwargs,
    ) -> None:
        super().__init__(
            gamma,
            max_flow,
            corr_radius,
            feature_channels,
            num_scales,
            upsample_factor,
            reg_refine,
            num_transformer_layers,
            num_head,
            ffn_dim_expansion,
            pred_bidir_flow,
            num_reg_refine,
            attn_type,
            attn_splits_list,
            corr_radius_list,
            prop_radius_list,
            **kwargs,
        )


class UniMatchScale2With6Refinements(UniMatch):
    pretrained_checkpoints = {
        "mix": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/unimatch_scale2_refine6-mixdata-398760b1.ckpt",
        "things": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/unimatch_scale2_refine6-things-54d7505b.ckpt",
        "sintel": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/unimatch_scale2_refine6-sintel-95ab1410.ckpt",
        "kitti": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/unimatch_scale2_refine6-kitti-0626279a.ckpt",
    }

    def __init__(
        self,
        gamma: float = 0.9,
        max_flow: float = 400,
        corr_radius: int = 4,
        feature_channels: int = 128,
        num_scales: int = 2,
        upsample_factor: int = 4,
        reg_refine: bool = True,
        num_transformer_layers: int = 6,
        num_head: int = 1,
        ffn_dim_expansion: int = 4,
        pred_bidir_flow: bool = False,
        num_reg_refine: int = 6,
        attn_type: str = "swin",
        attn_splits_list: Sequence[int] = (2, 8),
        corr_radius_list: Sequence[int] = (-1, 4),
        prop_radius_list: Sequence[int] = (-1, 1),
        **kwargs,
    ) -> None:
        super().__init__(
            gamma,
            max_flow,
            corr_radius,
            feature_channels,
            num_scales,
            upsample_factor,
            reg_refine,
            num_transformer_layers,
            num_head,
            ffn_dim_expansion,
            pred_bidir_flow,
            num_reg_refine,
            attn_type,
            attn_splits_list,
            corr_radius_list,
            prop_radius_list,
            **kwargs,
        )


@register_model
@trainable
class unimatch(UniMatch):
    pass


@register_model
@trainable
class unimatch_sc2(UniMatchScale2):
    pass


@register_model
@trainable
class unimatch_sc2_ref6(UniMatchScale2With6Refinements):
    pass


@register_model
@trainable
class gmflow_p(UniMatch):
    pass


@register_model
@trainable
class gmflow_p_sc2(UniMatchScale2):
    pass


@register_model
@trainable
class gmflow_p_sc2_ref6(UniMatchScale2With6Refinements):
    pass



from einops import rearrange
import torch
import torch.nn as nn
from torch.nn import init

from .flownetc import FlowNetC
from .flownets import FlowNetS
from .submodules import *
from .flownet_base import FlowNetBase


from ptlflow.utils.registry import register_model, trainable


class FlowNetCS(FlowNetBase):
    pretrained_checkpoints = {
        "things": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/flownetcs-things-4bdecffa.ckpt"
    }

    def __init__(
        self,
        div_flow: float = 20.0,
        input_channels: int = 12,
        batch_norm: bool = False,
        loss_start_scale: int = 4,
        loss_num_scales: int = 5,
        loss_base_weight: float = 0.32,
        loss_norm: str = "L2",
        **kwargs,
    ):
        super(FlowNetCS, self).__init__(
            div_flow=div_flow,
            input_channels=input_channels,
            batch_norm=batch_norm,
            loss_start_scale=loss_start_scale,
            loss_num_scales=loss_num_scales,
            loss_base_weight=loss_base_weight,
            loss_norm=loss_norm,
            **kwargs,
        )

        self.rgb_max = 1

        # First Block (FlowNetC)
        self.flownetc = FlowNetC(
            div_flow=div_flow,
            input_channels=input_channels,
            batch_norm=batch_norm,
            loss_start_scale=loss_start_scale,
            loss_num_scales=loss_num_scales,
            loss_base_weight=loss_base_weight,
            loss_norm=loss_norm,
            **kwargs,
        )

        # Block (FlowNetS)
        self.flownets_1 = FlowNetS(
            div_flow=div_flow,
            input_channels=input_channels,
            batch_norm=batch_norm,
            loss_start_scale=loss_start_scale,
            loss_num_scales=loss_num_scales,
            loss_base_weight=loss_base_weight,
            loss_norm=loss_norm,
            **kwargs,
        )

        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                if m.bias is not None:
                    init.uniform_(m.bias)
                init.xavier_uniform_(m.weight)

            if isinstance(m, nn.ConvTranspose2d):
                if m.bias is not None:
                    init.uniform_(m.bias)
                init.xavier_uniform_(m.weight)
                # init_deconv_bilinear(m.weight)

    def warp(self, x, flo):
        """
        warp an image/tensor (im2) back to im1, according to the optical flow
        x: [B, C, H, W] (im2)
        flo: [B, 2, H, W] flow
        """
        B, C, H, W = x.size()
        # mesh grid
        xx = torch.arange(0, W).view(1, -1).repeat(H, 1)
        yy = torch.arange(0, H).view(-1, 1).repeat(1, W)
        xx = xx.view(1, 1, H, W).repeat(B, 1, 1, 1)
        yy = yy.view(1, 1, H, W).repeat(B, 1, 1, 1)
        grid = torch.cat((xx, yy), 1)

        if x.is_cuda:
            grid = grid.to(dtype=x.dtype, device=x.device)
        vgrid = grid + flo

        # scale grid to [-1,1]
        vgrid[:, 0, :, :] = 2.0 * vgrid[:, 0, :, :].clone() / max(W - 1, 1) - 1.0
        vgrid[:, 1, :, :] = 2.0 * vgrid[:, 1, :, :].clone() / max(H - 1, 1) - 1.0

        vgrid = vgrid.permute(0, 2, 3, 1)
        output = nn.functional.grid_sample(x, vgrid, align_corners=True)
        mask = torch.ones(x.size()).to(dtype=x.dtype, device=x.device)
        mask = nn.functional.grid_sample(mask, vgrid, align_corners=True)

        mask[mask < 0.9999] = 0
        mask[mask > 0] = 1

        return output * mask

    def forward(self, inputs):
        orig_images = inputs["images"]
        bgr_mean = rearrange(orig_images, "b n c h w -> b n c (h w)").mean(-1)
        bgr_mean = bgr_mean[..., None, None]
        images, image_resizer = self.preprocess_images(
            orig_images,
            bgr_add=-bgr_mean,
            bgr_mult=1.0,
            bgr_to_rgb=True,
            resize_mode="interpolation",
            interpolation_mode="bilinear",
            interpolation_align_corners=True,
        )
        inputs["images"] = images

        # flownetc
        flownetc_flow = self.flownetc(inputs, skip_preprocess=True)["flows"][:, 0]

        # warp img1 to img0; magnitude of diff between img0 and and warped_img1,
        resampled_img1 = self.warp(inputs["images"][:, 1], flownetc_flow)
        diff_img0 = inputs["images"][:, 0] - resampled_img1
        norm_diff_img0 = torch.norm(diff_img0, p=2, dim=1, keepdim=True)

        # concat img0, img1, img1->img0, flow, diff-mag ;
        concat1 = torch.cat(
            (
                inputs["images"][:, 0],
                inputs["images"][:, 1],
                resampled_img1,
                flownetc_flow / self.div_flow,
                norm_diff_img0,
            ),
            dim=1,
        )

        # flownets1
        flownets1_preds = self.flownets_1(
            {"images": concat1[:, None]}, skip_preprocess=True
        )
        flownets1_preds["flows"] = self.postprocess_predictions(
            flownets1_preds["flows"], image_resizer, is_flow=True
        )

        inputs["images"] = orig_images

        return flownets1_preds


@register_model
@trainable
class flownetcs(FlowNetCS):
    pass



"""
implementation of the PWC-DC network for optical flow estimation by Sun et al., 2018
Jinwei Gu and Zhile Ren
"""

from typing import Optional, Sequence

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.autograd import Variable
import os

os.environ["PYTHON_EGG_CACHE"] = "tmp/"  # a writable directory
import numpy as np
import math

from ptlflow.utils.registry import register_model, trainable
from .submodule import pspnet_s, pspnet, conv
from .conv4d import sepConv4d, butterfly4D
from ..base_model.base_model import BaseModel


class VCNLoss(nn.Module):
    def __init__(self, maxdisp, fac, train_batch_size):
        super(VCNLoss, self).__init__()
        self.maxdisp = maxdisp
        self.fac = fac
        self.bs = train_batch_size

        self.warpx = WarpModule()

    def forward(self, preds, inputs):
        flowl0 = inputs["flows"][:, 0].clone()
        mask = (
            (inputs["valids"][:, 0, 0] == 1)
            & (flowl0[:, 0].abs() < self.maxdisp)
            & (flowl0[:, 1].abs() < (self.maxdisp // self.fac))
        )
        mask = mask.detach().clone()

        im = inputs["images"]
        im = im.view(im.shape[0] * im.shape[1], im.shape[2], im.shape[3], im.shape[4])

        # out-of-range loss
        im_warp, _ = self.warpx(im[self.bs :], flowl0[:, :2])
        occ_mask = (im_warp - im[: self.bs]).norm(dim=1) > 0.3

        m = 64.0
        loss = 0
        ws = [0.25, 0.25, 0.25, 0.5, 1.0]
        for i in range(len(ws)):
            loss += (
                ws[i]
                * torch.norm((preds["flow_preds"][i] * m - flowl0[:, :2]), 0, 1)[
                    mask
                ].mean()
            )
            m /= 2

        im_size = flowl0.shape[2:4]

        m = 32.0
        up_flows = preds["up_flows_preds"]
        for i in range(len(up_flows)):
            up_flows[i] = (
                F.interpolate(
                    up_flows[i],
                    [im_size[0], im_size[1]],
                    mode="bilinear",
                    align_corners=False,
                )
                * m
            )
            m /= 2
        up_flows.insert(0, 0)

        oors = preds["oors_preds"]
        for i in range(len(oors)):
            oors[i] = F.interpolate(
                oors[i][:, np.newaxis],
                [im_size[0], im_size[1]],
                mode="bilinear",
                align_corners=False,
            )[:, 0]

        m = 64
        for i in range(len(oors)):
            loss += self.get_oor_loss(
                flowl0[:, :2] - up_flows[i],
                oors[i],
                (m * preds["flow_reg_maxs_preds"][i]),
                occ_mask,
            )
            m /= 2

        return loss

    def get_oor_loss(self, flowl0, oor3, maxdisp, occ_mask):
        """
        return out-of-range loss
        """
        oor3_gt = (flowl0.abs() > maxdisp).detach()  #  (8*self.md[3])
        oor3_gt = (((oor3_gt.sum(1) > 0) + occ_mask) > 0).to(
            dtype=flowl0.dtype
        )  # oor, or occluded
        weights = oor3_gt.sum().to(dtype=flowl0.dtype) / (
            oor3_gt.shape[0] * oor3_gt.shape[1] * oor3_gt.shape[2]
        )
        weights = oor3_gt * (1 - weights) + (1 - oor3_gt) * weights
        loss_oor3 = F.binary_cross_entropy_with_logits(
            oor3, oor3_gt, size_average=True, weight=weights
        )
        return loss_oor3


class flow_reg(nn.Module):
    """
    Soft winner-take-all that selects the most likely diplacement.
    Set ent=True to enable entropy output.
    Set maxdisp to adjust maximum allowed displacement towards one side.
        maxdisp=4 searches for a 9x9 region.
    Set fac to squeeze search window.
        maxdisp=4 and fac=2 gives search window of 9x5
    """

    def __init__(self, ent=False, maxdisp=int(4), fac=1):
        super(flow_reg, self).__init__()
        self.ent = ent
        self.md = maxdisp
        self.fac = fac
        self.truncated = True
        self.wsize = 3  # by default using truncation 7x7

        self.create_flow([1, 1, 1])

        self.pool3d = nn.MaxPool3d(
            (self.wsize * 2 + 1, self.wsize * 2 + 1, 1),
            stride=1,
            padding=(self.wsize, self.wsize, 0),
        )

    def create_flow(self, size):
        B, W, H = size
        flowrangey = range(-self.md, self.md + 1)
        flowrangex = range(-int(self.md // self.fac), int(self.md // self.fac) + 1)
        meshgrid = np.meshgrid(flowrangex, flowrangey)
        flowy = np.tile(
            np.reshape(
                meshgrid[0],
                [1, 2 * self.md + 1, 2 * int(self.md // self.fac) + 1, 1, 1],
            ),
            (B, 1, 1, H, W),
        )
        flowx = np.tile(
            np.reshape(
                meshgrid[1],
                [1, 2 * self.md + 1, 2 * int(self.md // self.fac) + 1, 1, 1],
            ),
            (B, 1, 1, H, W),
        )
        self.flowx = torch.Tensor(flowx)
        self.flowy = torch.Tensor(flowy)

    def forward(self, x):
        b, u, v, h, w = x.shape
        oldx = x

        if (
            b != self.flowx.shape[0]
            or h != self.flowx.shape[3]
            or w != self.flowx.shape[4]
        ):
            self.create_flow((b, w, h))

        if self.truncated:
            # truncated softmax
            x = x.view(b, u * v, h, w)

            idx = x.argmax(1)[:, np.newaxis]
            if x.is_cuda:
                mask = Variable(torch.cuda.HalfTensor(b, u * v, h, w)).fill_(0)
            else:
                mask = Variable(torch.FloatTensor(b, u * v, h, w)).fill_(0)
            mask.scatter_(1, idx, 1)
            mask = mask.view(b, 1, u, v, -1)
            mask = self.pool3d(mask)[:, 0].view(b, u, v, h, w)

            ninf = x.clone().fill_(-np.inf).view(b, u, v, h, w)
            x = torch.where(mask.bool(), oldx, ninf)
        else:
            self.wsize = (np.sqrt(u * v) - 1) / 2

        b, u, v, h, w = x.shape
        x = F.softmax(x.view(b, -1, h, w), 1).view(b, u, v, h, w)
        outx = torch.sum(
            torch.sum(x * self.flowx.to(dtype=x.dtype, device=x.device), 1),
            1,
            keepdim=True,
        )
        outy = torch.sum(
            torch.sum(x * self.flowy.to(dtype=x.dtype, device=x.device), 1),
            1,
            keepdim=True,
        )

        if self.ent:
            # local
            local_entropy = (
                (-x * torch.clamp(x, 1e-9, 1 - 1e-9).log()).sum(1).sum(1)[:, np.newaxis]
            )
            if self.wsize == 0:
                local_entropy[:] = 1.0
            else:
                local_entropy /= np.log((self.wsize * 2 + 1) ** 2)

            # global
            x = F.softmax(oldx.view(b, -1, h, w), 1).view(b, u, v, h, w)
            global_entropy = (
                (-x * torch.clamp(x, 1e-9, 1 - 1e-9).log()).sum(1).sum(1)[:, np.newaxis]
            )
            global_entropy /= np.log(x.shape[1] * x.shape[2])
            return torch.cat([outx, outy], 1), torch.cat(
                [local_entropy, global_entropy], 1
            )
        else:
            return torch.cat([outx, outy], 1), None


class WarpModule(nn.Module):
    """
    taken from https://github.com/NVlabs/PWC-Net/blob/master/PyTorch/models/PWCNet.py
    """

    def __init__(self):
        super(WarpModule, self).__init__()
        self.create_grid([1, 1, 1])

    def create_grid(self, size):
        B, W, H = size
        # mesh grid
        xx = torch.arange(0, W).view(1, -1).repeat(H, 1)
        yy = torch.arange(0, H).view(-1, 1).repeat(1, W)
        xx = xx.view(1, 1, H, W).repeat(B, 1, 1, 1)
        yy = yy.view(1, 1, H, W).repeat(B, 1, 1, 1)

        # self.register_buffer('grid',torch.cat((xx,yy),1).float())
        self.grid = torch.cat((xx, yy), 1).float()

    def forward(self, x, flo):
        """
        warp an image/tensor (im2) back to im1, according to the optical flow
        x: [B, C, H, W] (im2)
        flo: [B, 2, H, W] flow
        """
        B, C, H, W = x.size()
        if (
            B != self.grid.shape[0]
            or H != self.grid.shape[2]
            or W != self.grid.shape[3]
        ):
            self.create_grid((B, W, H))

        vgrid = self.grid.to(dtype=flo.dtype, device=flo.device) + flo

        # scale grid to [-1,1]
        vgrid[:, 0, :, :] = 2.0 * vgrid[:, 0, :, :] / max(W - 1, 1) - 1.0
        vgrid[:, 1, :, :] = 2.0 * vgrid[:, 1, :, :] / max(H - 1, 1) - 1.0

        vgrid = vgrid.permute(0, 2, 3, 1)
        output = nn.functional.grid_sample(x, vgrid, align_corners=True)
        mask = ((vgrid[:, :, :, 0].abs() < 1) * (vgrid[:, :, :, 1].abs() < 1)) > 0
        return output * mask.unsqueeze(1), mask


class VCNSmall(BaseModel):
    """
    VCN.
    md defines maximum displacement for each level, following a coarse-to-fine-warping scheme
    fac defines squeeze parameter for the coarsest level
    """

    pretrained_checkpoints = {
        "chairs": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/vcn_small-chairs-56243688.ckpt",
        "things": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/vcn_small-things-2f19af2d.ckpt",
    }

    def __init__(
        self,
        fac: float = 1.0,
        maxdisp: int = 256,
        md: Sequence[int] = [4, 4, 4, 4, 4],
        train_batch_size: Optional[int] = None,
        **kwargs,
    ):
        super(VCNSmall, self).__init__(
            loss_fn=VCNLoss(maxdisp, fac, train_batch_size),
            output_stride=64,
            **kwargs,
        )
        self.md = md
        self.fac = fac
        self.maxdisp = maxdisp
        use_entropy = True
        withbn = True

        ## pspnet
        self.pspnet = pspnet_s(is_proj=False)

        ## Volumetric-UNet
        fdima1 = 128  # 6/5/4
        fdima2 = 64  # 3/2
        fdimb1 = 16  # 6/5/4/3
        fdimb2 = 12  # 2

        full = False
        self.f6 = butterfly4D(fdima1, fdimb1, withbn=withbn, full=full)
        self.p6 = sepConv4d(fdimb1, fdimb1, with_bn=False, full=full)

        self.f5 = butterfly4D(fdima1, fdimb1, withbn=withbn, full=full)
        self.p5 = sepConv4d(fdimb1, fdimb1, with_bn=False, full=full)

        self.f4 = butterfly4D(fdima1, fdimb1, withbn=withbn, full=full)
        self.p4 = sepConv4d(fdimb1, fdimb1, with_bn=False, full=full)

        self.f3 = butterfly4D(fdima2, fdimb1, withbn=withbn, full=full)
        self.p3 = sepConv4d(fdimb1, fdimb1, with_bn=False, full=full)

        ## soft WTA modules
        self.flow_reg64 = flow_reg(ent=use_entropy, maxdisp=self.md[0], fac=self.fac)
        self.flow_reg32 = flow_reg(ent=use_entropy, maxdisp=self.md[1])
        self.flow_reg16 = flow_reg(ent=use_entropy, maxdisp=self.md[2])
        self.flow_reg8 = flow_reg(ent=use_entropy, maxdisp=self.md[3])

        ## warping modules
        self.warp5 = WarpModule()
        self.warp4 = WarpModule()
        self.warp3 = WarpModule()
        if self.training:
            self.warpx = WarpModule()

        ## hypotheses fusion modules, adopted from the refinement module of PWCNet
        # https://github.com/NVlabs/PWC-Net/blob/master/PyTorch/models/PWCNet.py
        # c6
        self.dc6_conv1 = conv(
            128 + 4 * fdimb1, 128, kernel_size=3, stride=1, padding=1, dilation=1
        )
        self.dc6_conv2 = conv(128, 128, kernel_size=3, stride=1, padding=2, dilation=2)
        self.dc6_conv3 = conv(128, 128, kernel_size=3, stride=1, padding=4, dilation=4)
        self.dc6_conv4 = conv(128, 96, kernel_size=3, stride=1, padding=8, dilation=8)
        self.dc6_conv5 = conv(96, 64, kernel_size=3, stride=1, padding=16, dilation=16)
        self.dc6_conv6 = conv(64, 32, kernel_size=3, stride=1, padding=1, dilation=1)
        self.dc6_conv7 = nn.Conv2d(
            32, 2 * fdimb1, kernel_size=3, stride=1, padding=1, bias=True
        )

        # c5
        self.dc5_conv1 = conv(
            128 + 4 * fdimb1 * 2, 128, kernel_size=3, stride=1, padding=1, dilation=1
        )
        self.dc5_conv2 = conv(128, 128, kernel_size=3, stride=1, padding=2, dilation=2)
        self.dc5_conv3 = conv(128, 128, kernel_size=3, stride=1, padding=4, dilation=4)
        self.dc5_conv4 = conv(128, 96, kernel_size=3, stride=1, padding=8, dilation=8)
        self.dc5_conv5 = conv(96, 64, kernel_size=3, stride=1, padding=16, dilation=16)
        self.dc5_conv6 = conv(64, 32, kernel_size=3, stride=1, padding=1, dilation=1)
        self.dc5_conv7 = nn.Conv2d(
            32, 2 * fdimb1 * 2, kernel_size=3, stride=1, padding=1, bias=True
        )

        # c4
        self.dc4_conv1 = conv(
            128 + 4 * fdimb1 * 3, 128, kernel_size=3, stride=1, padding=1, dilation=1
        )
        self.dc4_conv2 = conv(128, 128, kernel_size=3, stride=1, padding=2, dilation=2)
        self.dc4_conv3 = conv(128, 128, kernel_size=3, stride=1, padding=4, dilation=4)
        self.dc4_conv4 = conv(128, 96, kernel_size=3, stride=1, padding=8, dilation=8)
        self.dc4_conv5 = conv(96, 64, kernel_size=3, stride=1, padding=16, dilation=16)
        self.dc4_conv6 = conv(64, 32, kernel_size=3, stride=1, padding=1, dilation=1)
        self.dc4_conv7 = nn.Conv2d(
            32, 2 * fdimb1 * 3, kernel_size=3, stride=1, padding=1, bias=True
        )

        # c3
        self.dc3_conv1 = conv(
            64 + 16 * fdimb1, 128, kernel_size=3, stride=1, padding=1, dilation=1
        )
        self.dc3_conv2 = conv(128, 128, kernel_size=3, stride=1, padding=2, dilation=2)
        self.dc3_conv3 = conv(128, 128, kernel_size=3, stride=1, padding=4, dilation=4)
        self.dc3_conv4 = conv(128, 96, kernel_size=3, stride=1, padding=8, dilation=8)
        self.dc3_conv5 = conv(96, 64, kernel_size=3, stride=1, padding=16, dilation=16)
        self.dc3_conv6 = conv(64, 32, kernel_size=3, stride=1, padding=1, dilation=1)
        self.dc3_conv7 = nn.Conv2d(
            32, 8 * fdimb1, kernel_size=3, stride=1, padding=1, bias=True
        )

        ## Out-of-range detection
        # if size[0]>1:  # only in train mode
        self.dc6_convo = nn.Sequential(
            conv(128 + 4 * fdimb1, 128, kernel_size=3, stride=1, padding=1, dilation=1),
            conv(128, 128, kernel_size=3, stride=1, padding=2, dilation=2),
            conv(128, 128, kernel_size=3, stride=1, padding=4, dilation=4),
            conv(128, 96, kernel_size=3, stride=1, padding=8, dilation=8),
            conv(96, 64, kernel_size=3, stride=1, padding=16, dilation=16),
            conv(64, 32, kernel_size=3, stride=1, padding=1, dilation=1),
            nn.Conv2d(32, 1, kernel_size=3, stride=1, padding=1, bias=True),
        )

        self.dc5_convo = nn.Sequential(
            conv(
                128 + 2 * 4 * fdimb1,
                128,
                kernel_size=3,
                stride=1,
                padding=1,
                dilation=1,
            ),
            conv(128, 128, kernel_size=3, stride=1, padding=2, dilation=2),
            conv(128, 128, kernel_size=3, stride=1, padding=4, dilation=4),
            conv(128, 96, kernel_size=3, stride=1, padding=8, dilation=8),
            conv(96, 64, kernel_size=3, stride=1, padding=16, dilation=16),
            conv(64, 32, kernel_size=3, stride=1, padding=1, dilation=1),
            nn.Conv2d(32, 1, kernel_size=3, stride=1, padding=1, bias=True),
        )

        self.dc4_convo = nn.Sequential(
            conv(
                128 + 3 * 4 * fdimb1,
                128,
                kernel_size=3,
                stride=1,
                padding=1,
                dilation=1,
            ),
            conv(128, 128, kernel_size=3, stride=1, padding=2, dilation=2),
            conv(128, 128, kernel_size=3, stride=1, padding=4, dilation=4),
            conv(128, 96, kernel_size=3, stride=1, padding=8, dilation=8),
            conv(96, 64, kernel_size=3, stride=1, padding=16, dilation=16),
            conv(64, 32, kernel_size=3, stride=1, padding=1, dilation=1),
            nn.Conv2d(32, 1, kernel_size=3, stride=1, padding=1, bias=True),
        )

        self.dc3_convo = nn.Sequential(
            conv(64 + 16 * fdimb1, 128, kernel_size=3, stride=1, padding=1, dilation=1),
            conv(128, 128, kernel_size=3, stride=1, padding=2, dilation=2),
            conv(128, 128, kernel_size=3, stride=1, padding=4, dilation=4),
            conv(128, 96, kernel_size=3, stride=1, padding=8, dilation=8),
            conv(96, 64, kernel_size=3, stride=1, padding=16, dilation=16),
            conv(64, 32, kernel_size=3, stride=1, padding=1, dilation=1),
            nn.Conv2d(32, 1, kernel_size=3, stride=1, padding=1, bias=True),
        )

        for m in self.modules():
            if isinstance(m, nn.Conv3d):
                n = (
                    m.kernel_size[0]
                    * m.kernel_size[1]
                    * m.kernel_size[2]
                    * m.out_channels
                )
                m.weight.data.normal_(0, math.sqrt(2.0 / n))
                if hasattr(m.bias, "data"):
                    m.bias.data.zero_()

    # @profile
    def corr(self, refimg_fea, targetimg_fea, maxdisp, fac=1):
        """
        correlation function. Adopted from https://github.com/ClementPinard/Pytorch-Correlation-extension
        faster, but backwards not implemented.
        """
        from spatial_correlation_sampler import SpatialCorrelationSampler

        corr = SpatialCorrelationSampler(
            kernel_size=1,
            patch_size=(int(1 + 2 * maxdisp // fac), int(1 + 2 * maxdisp)),
            stride=1,
            padding=0,
            dilation_patch=1,
        )
        cost = corr(refimg_fea, targetimg_fea)
        cost = F.leaky_relu(cost, 0.1, inplace=True)
        return cost

    def corrf(self, refimg_fea, targetimg_fea, maxdisp, fac=1):
        """
        another correlation function giving the same result as corr()
        supports backwards
        """
        b, c, height, width = refimg_fea.shape
        tensor_func = (
            torch.FloatTensor if refimg_fea.dtype == torch.float32 else torch.HalfTensor
        )
        cost = (
            Variable(
                tensor_func(
                    b, c, 2 * maxdisp + 1, 2 * int(maxdisp // fac) + 1, height, width
                )
            )
            .fill_(0.0)
            .to(refimg_fea.device)
        )  # b,c,u,v,h,w
        for i in range(2 * maxdisp + 1):
            ind = i - maxdisp
            for j in range(2 * int(maxdisp // fac) + 1):
                indd = j - int(maxdisp // fac)
                feata = refimg_fea[
                    :, :, max(0, -indd) : height - indd, max(0, -ind) : width - ind
                ]
                featb = targetimg_fea[
                    :, :, max(0, +indd) : height + indd, max(0, ind) : width + ind
                ]
                diff = feata * featb
                cost[
                    :,
                    :,
                    i,
                    j,
                    max(0, -indd) : height - indd,
                    max(0, -ind) : width - ind,
                ] = diff
        cost = F.leaky_relu(cost, 0.1, inplace=True)
        return cost

    def get_oor_loss(self, flowl0, oor3, maxdisp, occ_mask):
        """
        return out-of-range loss
        """
        oor3_gt = (flowl0.abs() > maxdisp).detach()  #  (8*self.md[3])
        oor3_gt = (((oor3_gt.sum(1) > 0) + occ_mask) > 0).to(
            dtype=flowl0.dtype
        )  # oor, or occluded
        weights = oor3_gt.sum().to(dtype=flowl0.dtype) / (
            oor3_gt.shape[0] * oor3_gt.shape[1] * oor3_gt.shape[2]
        )
        weights = oor3_gt * (1 - weights) + (1 - oor3_gt) * weights
        loss_oor3 = F.binary_cross_entropy_with_logits(
            oor3, oor3_gt, size_average=True, weight=weights
        )
        return loss_oor3

    def weight_parameters(self):
        return [param for name, param in self.named_parameters() if "weight" in name]

    def bias_parameters(self):
        return [param for name, param in self.named_parameters() if "bias" in name]

    def forward(self, inputs, disc_aux=None):
        if self.extra_params is not None and "mean_bgr_L" in self.extra_params:
            mean_bgr = self.extra_params["mean_bgr_L"]
        else:
            mean_bgr = [0.33, 0.33, 0.33]
        mean_bgr = [-x for x in mean_bgr]

        im, image_resizer = self.preprocess_images(
            inputs["images"],
            bgr_add=mean_bgr,
            bgr_mult=1.0,
            bgr_to_rgb=True,
            resize_mode="interpolation",
            interpolation_mode="bilinear",
            interpolation_align_corners=True,
        )
        im = im.view(im.shape[0] * im.shape[1], im.shape[2], im.shape[3], im.shape[4])

        bs = im.shape[0] // 2

        c06, c05, c04, c03 = self.pspnet(im)
        c16 = c06[:bs]
        c26 = c06[bs:]
        c15 = c05[:bs]
        c25 = c05[bs:]
        c14 = c04[:bs]
        c24 = c04[bs:]
        c13 = c03[:bs]
        c23 = c03[bs:]

        # normalize
        c16n = c16 / (c16.norm(dim=1, keepdim=True) + 1e-9)
        c26n = c26 / (c26.norm(dim=1, keepdim=True) + 1e-9)
        c15n = c15 / (c15.norm(dim=1, keepdim=True) + 1e-9)
        c25n = c25 / (c25.norm(dim=1, keepdim=True) + 1e-9)
        c14n = c14 / (c14.norm(dim=1, keepdim=True) + 1e-9)
        c24n = c24 / (c24.norm(dim=1, keepdim=True) + 1e-9)
        c13n = c13 / (c13.norm(dim=1, keepdim=True) + 1e-9)
        c23n = c23 / (c23.norm(dim=1, keepdim=True) + 1e-9)

        up_flows = []
        oors = []
        flow_reg_maxs = []

        ## matching 6
        if self.training or (not im.is_cuda):
            feat6 = self.corrf(c16n, c26n, self.md[0], fac=self.fac)
        else:
            feat6 = self.corrf(c16n, c26n, self.md[0], fac=self.fac)
            # feat6 = self.corr(c16n,c26n,self.md[0],fac=self.fac)
        feat6 = self.f6(feat6)
        cost6 = self.p6(feat6)  # b, 16, u,v,h,w

        # soft WTA
        b, c, u, v, h, w = cost6.shape
        cost6 = cost6.view(
            -1, u, v, h, w
        )  # bx16, 9,9,h,w, also predict uncertainty from here
        flow6h, ent6h = self.flow_reg64(cost6)  # bx16, 2, h, w
        flow6h = flow6h.view(bs, -1, h, w)  # b, 16*2, h, w
        ent6h = ent6h.view(bs, -1, h, w)  # b, 16*1, h, w
        flow_reg_maxs.append(self.flow_reg64.flowx.max())

        if self.training:
            x = torch.cat((ent6h.detach(), flow6h.detach(), c16), 1)
            oor6 = self.dc6_convo(x)[:, 0]
            oors.append(oor6)

        # hypotheses fusion
        x = torch.cat((ent6h.detach(), flow6h.detach(), c16), 1)
        x = self.dc6_conv4(self.dc6_conv3(self.dc6_conv2(self.dc6_conv1(x))))
        va = self.dc6_conv7(self.dc6_conv6(self.dc6_conv5(x)))
        va = va.view(b, -1, 2, h, w)
        flow6 = (flow6h.view(b, -1, 2, h, w) * F.softmax(va, 1)).sum(1)

        ## matching 5
        up_flow6 = (
            F.interpolate(
                flow6,
                [im.size()[2] // 32, im.size()[3] // 32],
                mode="bilinear",
                align_corners=False,
            )
            * 2
        )
        up_flows.append(up_flow6)
        warp5, _ = self.warp5(c25n, up_flow6)
        if self.training or (not im.is_cuda):
            feat5 = self.corrf(c15n, warp5, self.md[1])
        else:
            feat5 = self.corrf(c15n, warp5, self.md[1])
            # feat5 = self.corr(c15n,warp5,self.md[1])
        feat5 = self.f5(feat5)
        cost5 = self.p5(feat5)  # b, 16, u,v,h,w

        # soft WTA
        b, c, u, v, h, w = cost5.shape
        cost5 = cost5.view(
            -1, u, v, h, w
        )  # bx16, 9,9,h,w, also predict uncertainty from here
        flow5h, ent5h = self.flow_reg32(cost5)  # bx16, 2, h, w
        flow5h = flow5h.view(b, c, 2, h, w) + up_flow6[:, np.newaxis]
        flow5h = flow5h.view(bs, -1, h, w)  # b, 16*2, h, w
        ent5h = ent5h.view(bs, -1, h, w)  # b, 16*1, h, w
        flow_reg_maxs.append(self.flow_reg32.flowx.max())

        # append coarse hypotheses
        flow5h = torch.cat(
            (
                flow5h,
                F.interpolate(
                    flow6h.detach() * 2,
                    [flow5h.shape[2], flow5h.shape[3]],
                    mode="bilinear",
                    align_corners=False,
                ),
            ),
            1,
        )  # b, k2--k2, h, w
        ent5h = torch.cat(
            (
                ent5h,
                F.interpolate(
                    ent6h,
                    [flow5h.shape[2], flow5h.shape[3]],
                    mode="bilinear",
                    align_corners=False,
                ),
            ),
            1,
        )

        if self.training:
            x = torch.cat((ent5h.detach(), flow5h.detach(), c15), 1)
            oor5 = self.dc5_convo(x)[:, 0]
            oors.append(oor5)

        # hypotheses fusion
        x = torch.cat((ent5h.detach(), flow5h.detach(), c15), 1)
        x = self.dc5_conv4(self.dc5_conv3(self.dc5_conv2(self.dc5_conv1(x))))
        va5 = self.dc5_conv7(self.dc5_conv6(self.dc5_conv5(x)))
        va5 = va5.view(b, -1, 2, h, w)
        flow5 = (flow5h.view(b, -1, 2, h, w) * F.softmax(va5, 1)).sum(
            1
        )  # b, 2k, 2, h, w

        ## matching 4
        up_flow5 = (
            F.interpolate(
                flow5,
                [im.size()[2] // 16, im.size()[3] // 16],
                mode="bilinear",
                align_corners=False,
            )
            * 2
        )
        up_flows.append(up_flow5)
        warp4, _ = self.warp4(c24n, up_flow5)
        if self.training or (not im.is_cuda):
            feat4 = self.corrf(c14n, warp4, self.md[2])
        else:
            feat4 = self.corrf(c14n, warp4, self.md[2])
            # feat4 = self.corr(c14n,warp4,self.md[2])
        feat4 = self.f4(feat4)
        cost4 = self.p4(feat4)  # b, 16, u,v,h,w

        # soft WTA
        b, c, u, v, h, w = cost4.shape
        cost4 = cost4.view(
            -1, u, v, h, w
        )  # bx16, 9,9,h,w, also predict uncertainty from here
        flow4h, ent4h = self.flow_reg16(cost4)  # bx16, 2, h, w
        flow4h = flow4h.view(b, c, 2, h, w) + up_flow5[:, np.newaxis]
        flow4h = flow4h.view(bs, -1, h, w)  # b, 16*2, h, w
        ent4h = ent4h.view(bs, -1, h, w)  # b, 16*1, h, w
        flow_reg_maxs.append(self.flow_reg16.flowx.max())

        # append coarse hypotheses
        flow4h = torch.cat(
            (
                flow4h,
                F.interpolate(
                    flow5h.detach() * 2,
                    [flow4h.shape[2], flow4h.shape[3]],
                    mode="bilinear",
                    align_corners=False,
                ),
            ),
            1,
        )
        ent4h = torch.cat(
            (
                ent4h,
                F.interpolate(
                    ent5h,
                    [flow4h.shape[2], flow4h.shape[3]],
                    mode="bilinear",
                    align_corners=False,
                ),
            ),
            1,
        )

        if self.training:
            x = torch.cat((ent4h.detach(), flow4h.detach(), c14), 1)
            oor4 = self.dc4_convo(x)[:, 0]
            oors.append(oor4)

        # hypotheses fusion
        x = torch.cat((ent4h.detach(), flow4h.detach(), c14), 1)
        x = self.dc4_conv4(self.dc4_conv3(self.dc4_conv2(self.dc4_conv1(x))))
        va = self.dc4_conv7(self.dc4_conv6(self.dc4_conv5(x)))
        va = va.view(b, -1, 2, h, w)
        flow4 = (flow4h.view(b, -1, 2, h, w) * F.softmax(va, 1)).sum(1)

        ## matching 3
        up_flow4 = (
            F.interpolate(
                flow4,
                [im.size()[2] // 8, im.size()[3] // 8],
                mode="bilinear",
                align_corners=False,
            )
            * 2
        )
        up_flows.append(up_flow4)
        warp3, _ = self.warp3(c23n, up_flow4)
        if self.training or (not im.is_cuda):
            feat3 = self.corrf(c13n, warp3, self.md[3])
        else:
            feat3 = self.corrf(c13n, warp3, self.md[3])
            # feat3 = self.corr(c13n,warp3,self.md[3])
        feat3 = self.f3(feat3)
        cost3 = self.p3(feat3)  # b, 16, u,v,h,w

        # soft WTA
        b, c, u, v, h, w = cost3.shape
        cost3 = cost3.view(
            -1, u, v, h, w
        )  # bx16, 9,9,h,w, also predict uncertainty from here
        flow3h, ent3h = self.flow_reg8(cost3)  # bx16, 2, h, w
        flow3h = flow3h.view(b, c, 2, h, w) + up_flow4[:, np.newaxis]
        flow3h = flow3h.view(bs, -1, h, w)  # b, 16*2, h, w
        ent3h = ent3h.view(bs, -1, h, w)  # b, 16*1, h, w
        flow_reg_maxs.append(self.flow_reg8.flowx.max())

        # append coarse hypotheses
        flow3h = torch.cat(
            (
                flow3h,
                F.interpolate(
                    flow4h.detach() * 2,
                    [flow3h.shape[2], flow3h.shape[3]],
                    mode="bilinear",
                    align_corners=False,
                ),
            ),
            1,
        )
        ent3h = torch.cat(
            (
                ent3h,
                F.interpolate(
                    ent4h,
                    [flow3h.shape[2], flow3h.shape[3]],
                    mode="bilinear",
                    align_corners=False,
                ),
            ),
            1,
        )

        if self.training:
            x = torch.cat((ent3h.detach(), flow3h.detach(), c13), 1)
            oor3 = self.dc3_convo(x)[:, 0]
            oors.append(oor3)

        # hypotheses fusion
        x = torch.cat((ent3h.detach(), flow3h.detach(), c13), 1)
        x = self.dc3_conv4(self.dc3_conv3(self.dc3_conv2(self.dc3_conv1(x))))
        va = self.dc3_conv7(self.dc3_conv6(self.dc3_conv5(x)))
        va = va.view(b, -1, 2, h, w)
        flow3 = (flow3h.view(b, -1, 2, h, w) * F.softmax(va, 1)).sum(1)

        flow3 = F.interpolate(
            flow3, [im.size()[2], im.size()[3]], mode="bilinear", align_corners=False
        )
        flow3 = self.postprocess_predictions(flow3, image_resizer, is_flow=True)
        flow4 = F.interpolate(
            flow4, [im.size()[2], im.size()[3]], mode="bilinear", align_corners=False
        )
        flow4 = self.postprocess_predictions(flow4, image_resizer, is_flow=True)
        flow5 = F.interpolate(
            flow5, [im.size()[2], im.size()[3]], mode="bilinear", align_corners=False
        )
        flow5 = self.postprocess_predictions(flow5, image_resizer, is_flow=True)
        flow6 = F.interpolate(
            flow6, [im.size()[2], im.size()[3]], mode="bilinear", align_corners=False
        )
        flow6 = self.postprocess_predictions(flow6, image_resizer, is_flow=True)
        flow2 = flow3 * 2
        flows = [flow6, flow5, flow4, flow3, flow2]

        outputs = {}

        if self.training:
            outputs["flow_preds"] = flows
            outputs["up_flows_preds"] = up_flows
            outputs["oors_preds"] = oors
            outputs["flow_reg_maxs_preds"] = flow_reg_maxs
            outputs["flows"] = flow2[:, None] * 4
        else:
            outputs["flows"] = flow2[:, None] * 4

        return outputs


class VCN(VCNSmall):
    """
    VCN.
    md defines maximum displacement for each level, following a coarse-to-fine-warping scheme
    fac defines squeeze parameter for the coarsest level
    """

    pretrained_checkpoints = {
        "chairs": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/vcn-chairs-6d37848e.ckpt",
        "things": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/vcn-things-171a566b.ckpt",
        "sintel": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/vcn-sintel-a3e6d864.ckpt",
        "kitti": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/vcn-kitti-3f6d9f0b.ckpt",
    }

    def __init__(
        self,
        fac: float = 1.0,
        maxdisp: int = 256,
        md: Sequence[int] = [4, 4, 4, 4, 4],
        train_batch_size: Optional[int] = None,
        **kwargs,
    ):
        super(VCN, self).__init__(
            fac=fac,
            maxdisp=maxdisp,
            md=md,
            train_batch_size=train_batch_size,
            **kwargs,
        )
        use_entropy = True
        withbn = True

        ## pspnet
        self.pspnet = pspnet(is_proj=False)

        ## Volumetric-UNet
        fdima1 = 128  # 6/5/4
        fdima2 = 64  # 3/2
        fdimb1 = 16  # 6/5/4/3
        fdimb2 = 12  # 2

        full = True
        self.f2 = butterfly4D(fdima2, fdimb2, withbn=withbn, full=full)
        self.p2 = sepConv4d(fdimb2, fdimb2, with_bn=False, full=full)

        ## soft WTA modules
        self.flow_reg4 = flow_reg(ent=use_entropy, maxdisp=self.md[4])

        ## warping modules
        self.warp2 = WarpModule()

        ## hypotheses fusion modules, adopted from the refinement module of PWCNet
        # https://github.com/NVlabs/PWC-Net/blob/master/PyTorch/models/PWCNet.py
        # c2
        self.dc2_conv1 = conv(
            64 + 16 * fdimb1 + 4 * fdimb2,
            128,
            kernel_size=3,
            stride=1,
            padding=1,
            dilation=1,
        )
        self.dc2_conv2 = conv(128, 128, kernel_size=3, stride=1, padding=2, dilation=2)
        self.dc2_conv3 = conv(128, 128, kernel_size=3, stride=1, padding=4, dilation=4)
        self.dc2_conv4 = conv(128, 96, kernel_size=3, stride=1, padding=8, dilation=8)
        self.dc2_conv5 = conv(96, 64, kernel_size=3, stride=1, padding=16, dilation=16)
        self.dc2_conv6 = conv(64, 32, kernel_size=3, stride=1, padding=1, dilation=1)
        self.dc2_conv7 = nn.Conv2d(
            32,
            4 * 2 * fdimb1 + 2 * fdimb2,
            kernel_size=3,
            stride=1,
            padding=1,
            bias=True,
        )

        ## Out-of-range detection
        # if size[0]>1:  # only in train mode
        self.dc2_convo = nn.Sequential(
            conv(
                64 + 16 * fdimb1 + 4 * fdimb2,
                128,
                kernel_size=3,
                stride=1,
                padding=1,
                dilation=1,
            ),
            conv(128, 128, kernel_size=3, stride=1, padding=2, dilation=2),
            conv(128, 128, kernel_size=3, stride=1, padding=4, dilation=4),
            conv(128, 96, kernel_size=3, stride=1, padding=8, dilation=8),
            conv(96, 64, kernel_size=3, stride=1, padding=16, dilation=16),
            conv(64, 32, kernel_size=3, stride=1, padding=1, dilation=1),
            nn.Conv2d(32, 1, kernel_size=3, stride=1, padding=1, bias=True),
        )

        for m in self.modules():
            if isinstance(m, nn.Conv3d):
                n = (
                    m.kernel_size[0]
                    * m.kernel_size[1]
                    * m.kernel_size[2]
                    * m.out_channels
                )
                m.weight.data.normal_(0, math.sqrt(2.0 / n))
                if hasattr(m.bias, "data"):
                    m.bias.data.zero_()

    def forward(self, inputs, disc_aux=None):
        # The original implementation uses different BGR means for im1 and im2
        # Here, we use the same for both to make it simpler
        if self.extra_params is not None and "mean_bgr_L" in self.extra_params:
            mean_bgr = self.extra_params["mean_bgr_L"]
        else:
            mean_bgr = [0.33, 0.33, 0.33]
        mean_bgr = [-x for x in mean_bgr]

        im, image_resizer = self.preprocess_images(
            inputs["images"],
            bgr_add=mean_bgr,
            bgr_mult=1.0,
            bgr_to_rgb=True,
            resize_mode="interpolation",
            interpolation_mode="bilinear",
            interpolation_align_corners=True,
        )

        im = im.view(im.shape[0] * im.shape[1], im.shape[2], im.shape[3], im.shape[4])

        bs = im.shape[0] // 2

        c06, c05, c04, c03, c02 = self.pspnet(im)
        c16 = c06[:bs]
        c26 = c06[bs:]
        c15 = c05[:bs]
        c25 = c05[bs:]
        c14 = c04[:bs]
        c24 = c04[bs:]
        c13 = c03[:bs]
        c23 = c03[bs:]
        c12 = c02[:bs]
        c22 = c02[bs:]

        # normalize
        c16n = c16 / (c16.norm(dim=1, keepdim=True) + 1e-9)
        c26n = c26 / (c26.norm(dim=1, keepdim=True) + 1e-9)
        c15n = c15 / (c15.norm(dim=1, keepdim=True) + 1e-9)
        c25n = c25 / (c25.norm(dim=1, keepdim=True) + 1e-9)
        c14n = c14 / (c14.norm(dim=1, keepdim=True) + 1e-9)
        c24n = c24 / (c24.norm(dim=1, keepdim=True) + 1e-9)
        c13n = c13 / (c13.norm(dim=1, keepdim=True) + 1e-9)
        c23n = c23 / (c23.norm(dim=1, keepdim=True) + 1e-9)
        c12n = c12 / (c12.norm(dim=1, keepdim=True) + 1e-9)
        c22n = c22 / (c22.norm(dim=1, keepdim=True) + 1e-9)

        up_flows = []
        oors = []
        flow_reg_maxs = []

        ## matching 6
        if self.training or (not im.is_cuda):
            feat6 = self.corrf(c16n, c26n, self.md[0], fac=self.fac)
        else:
            feat6 = self.corrf(c16n, c26n, self.md[0], fac=self.fac)
            # feat6 = self.corr(c16n,c26n,self.md[0],fac=self.fac)
        feat6 = self.f6(feat6)
        cost6 = self.p6(feat6)  # b, 16, u,v,h,w

        # soft WTA
        b, c, u, v, h, w = cost6.shape
        cost6 = cost6.view(
            -1, u, v, h, w
        )  # bx16, 9,9,h,w, also predict uncertainty from here
        flow6h, ent6h = self.flow_reg64(cost6)  # bx16, 2, h, w
        flow6h = flow6h.view(bs, -1, h, w)  # b, 16*2, h, w
        ent6h = ent6h.view(bs, -1, h, w)  # b, 16*1, h, w
        flow_reg_maxs.append(self.flow_reg64.flowx.max())

        if self.training:
            x = torch.cat((ent6h.detach(), flow6h.detach(), c16), 1)
            oor6 = self.dc6_convo(x)[:, 0]
            oors.append(oor6)

        # hypotheses fusion
        x = torch.cat((ent6h.detach(), flow6h.detach(), c16), 1)
        x = self.dc6_conv4(self.dc6_conv3(self.dc6_conv2(self.dc6_conv1(x))))
        va = self.dc6_conv7(self.dc6_conv6(self.dc6_conv5(x)))
        va = va.view(b, -1, 2, h, w)
        flow6 = (flow6h.view(b, -1, 2, h, w) * F.softmax(va, 1)).sum(1)

        ## matching 5
        up_flow6 = (
            F.interpolate(
                flow6,
                [im.size()[2] // 32, im.size()[3] // 32],
                mode="bilinear",
                align_corners=False,
            )
            * 2
        )
        up_flows.append(up_flow6)
        warp5, _ = self.warp5(c25n, up_flow6)
        if self.training or (not im.is_cuda):
            feat5 = self.corrf(c15n, warp5, self.md[1])
        else:
            feat5 = self.corrf(c15n, warp5, self.md[1])
            # feat5 = self.corr(c15n,warp5,self.md[1])
        feat5 = self.f5(feat5)
        cost5 = self.p5(feat5)  # b, 16, u,v,h,w

        # soft WTA
        b, c, u, v, h, w = cost5.shape
        cost5 = cost5.view(
            -1, u, v, h, w
        )  # bx16, 9,9,h,w, also predict uncertainty from here
        flow5h, ent5h = self.flow_reg32(cost5)  # bx16, 2, h, w
        flow5h = flow5h.view(b, c, 2, h, w) + up_flow6[:, np.newaxis]
        flow5h = flow5h.view(bs, -1, h, w)  # b, 16*2, h, w
        ent5h = ent5h.view(bs, -1, h, w)  # b, 16*1, h, w
        flow_reg_maxs.append(self.flow_reg32.flowx.max())

        # append coarse hypotheses
        flow5h = torch.cat(
            (
                flow5h,
                F.interpolate(
                    flow6h.detach() * 2,
                    [flow5h.shape[2], flow5h.shape[3]],
                    mode="bilinear",
                    align_corners=False,
                ),
            ),
            1,
        )  # b, k2--k2, h, w
        ent5h = torch.cat(
            (
                ent5h,
                F.interpolate(
                    ent6h,
                    [flow5h.shape[2], flow5h.shape[3]],
                    mode="bilinear",
                    align_corners=False,
                ),
            ),
            1,
        )

        if self.training:
            x = torch.cat((ent5h.detach(), flow5h.detach(), c15), 1)
            oor5 = self.dc5_convo(x)[:, 0]
            oors.append(oor5)

        # hypotheses fusion
        x = torch.cat((ent5h.detach(), flow5h.detach(), c15), 1)
        x = self.dc5_conv4(self.dc5_conv3(self.dc5_conv2(self.dc5_conv1(x))))
        va5 = self.dc5_conv7(self.dc5_conv6(self.dc5_conv5(x)))
        va5 = va5.view(b, -1, 2, h, w)
        flow5 = (flow5h.view(b, -1, 2, h, w) * F.softmax(va5, 1)).sum(
            1
        )  # b, 2k, 2, h, w

        ## matching 4
        up_flow5 = (
            F.interpolate(
                flow5,
                [im.size()[2] // 16, im.size()[3] // 16],
                mode="bilinear",
                align_corners=False,
            )
            * 2
        )
        up_flows.append(up_flow5)
        warp4, _ = self.warp4(c24n, up_flow5)
        if self.training or (not im.is_cuda):
            feat4 = self.corrf(c14n, warp4, self.md[2])
        else:
            feat4 = self.corrf(c14n, warp4, self.md[2])
            # feat4 = self.corr(c14n,warp4,self.md[2])
        feat4 = self.f4(feat4)
        cost4 = self.p4(feat4)  # b, 16, u,v,h,w

        # soft WTA
        b, c, u, v, h, w = cost4.shape
        cost4 = cost4.view(
            -1, u, v, h, w
        )  # bx16, 9,9,h,w, also predict uncertainty from here
        flow4h, ent4h = self.flow_reg16(cost4)  # bx16, 2, h, w
        flow4h = flow4h.view(b, c, 2, h, w) + up_flow5[:, np.newaxis]
        flow4h = flow4h.view(bs, -1, h, w)  # b, 16*2, h, w
        ent4h = ent4h.view(bs, -1, h, w)  # b, 16*1, h, w
        flow_reg_maxs.append(self.flow_reg16.flowx.max())

        # append coarse hypotheses
        flow4h = torch.cat(
            (
                flow4h,
                F.interpolate(
                    flow5h.detach() * 2,
                    [flow4h.shape[2], flow4h.shape[3]],
                    mode="bilinear",
                    align_corners=False,
                ),
            ),
            1,
        )
        ent4h = torch.cat(
            (
                ent4h,
                F.interpolate(
                    ent5h,
                    [flow4h.shape[2], flow4h.shape[3]],
                    mode="bilinear",
                    align_corners=False,
                ),
            ),
            1,
        )

        if self.training:
            x = torch.cat((ent4h.detach(), flow4h.detach(), c14), 1)
            oor4 = self.dc4_convo(x)[:, 0]
            oors.append(oor4)

        # hypotheses fusion
        x = torch.cat((ent4h.detach(), flow4h.detach(), c14), 1)
        x = self.dc4_conv4(self.dc4_conv3(self.dc4_conv2(self.dc4_conv1(x))))
        va = self.dc4_conv7(self.dc4_conv6(self.dc4_conv5(x)))
        va = va.view(b, -1, 2, h, w)
        flow4 = (flow4h.view(b, -1, 2, h, w) * F.softmax(va, 1)).sum(1)

        ## matching 3
        up_flow4 = (
            F.interpolate(
                flow4,
                [im.size()[2] // 8, im.size()[3] // 8],
                mode="bilinear",
                align_corners=False,
            )
            * 2
        )
        up_flows.append(up_flow4)
        warp3, _ = self.warp3(c23n, up_flow4)
        if self.training or (not im.is_cuda):
            feat3 = self.corrf(c13n, warp3, self.md[3])
        else:
            feat3 = self.corrf(c13n, warp3, self.md[3])
            # feat3 = self.corr(c13n,warp3,self.md[3])
        feat3 = self.f3(feat3)
        cost3 = self.p3(feat3)  # b, 16, u,v,h,w

        # soft WTA
        b, c, u, v, h, w = cost3.shape
        cost3 = cost3.view(
            -1, u, v, h, w
        )  # bx16, 9,9,h,w, also predict uncertainty from here
        flow3h, ent3h = self.flow_reg8(cost3)  # bx16, 2, h, w
        flow3h = flow3h.view(b, c, 2, h, w) + up_flow4[:, np.newaxis]
        flow3h = flow3h.view(bs, -1, h, w)  # b, 16*2, h, w
        ent3h = ent3h.view(bs, -1, h, w)  # b, 16*1, h, w
        flow_reg_maxs.append(self.flow_reg8.flowx.max())

        # append coarse hypotheses
        flow3h = torch.cat(
            (
                flow3h,
                F.interpolate(
                    flow4h.detach() * 2,
                    [flow3h.shape[2], flow3h.shape[3]],
                    mode="bilinear",
                    align_corners=False,
                ),
            ),
            1,
        )
        ent3h = torch.cat(
            (
                ent3h,
                F.interpolate(
                    ent4h,
                    [flow3h.shape[2], flow3h.shape[3]],
                    mode="bilinear",
                    align_corners=False,
                ),
            ),
            1,
        )

        if self.training:
            x = torch.cat((ent3h.detach(), flow3h.detach(), c13), 1)
            oor3 = self.dc3_convo(x)[:, 0]
            oors.append(oor3)

        # hypotheses fusion
        x = torch.cat((ent3h.detach(), flow3h.detach(), c13), 1)
        x = self.dc3_conv4(self.dc3_conv3(self.dc3_conv2(self.dc3_conv1(x))))
        va = self.dc3_conv7(self.dc3_conv6(self.dc3_conv5(x)))
        va = va.view(b, -1, 2, h, w)
        flow3 = (flow3h.view(b, -1, 2, h, w) * F.softmax(va, 1)).sum(1)

        ## matching 2
        up_flow3 = (
            F.interpolate(
                flow3,
                [im.size()[2] // 4, im.size()[3] // 4],
                mode="bilinear",
                align_corners=False,
            )
            * 2
        )
        up_flows.append(up_flow3)
        warp2, _ = self.warp2(c22n, up_flow3)
        if self.training or (not im.is_cuda):
            feat2 = self.corrf(c12n, warp2, self.md[4])
        else:
            feat2 = self.corrf(c12n, warp2, self.md[4])
            # feat2 = self.corr(c12n,warp2,self.md[4])
        feat2 = self.f2(feat2)
        cost2 = self.p2(feat2)  # b, 16, u,v,h,w

        # soft WTA
        b, c, u, v, h, w = cost2.shape
        cost2 = cost2.view(
            -1, u, v, h, w
        )  # bx12, 9,9,h,w, also predict uncertainty from here
        flow2h, ent2h = self.flow_reg4(cost2)  # bx12, 2, h, w
        flow2h = flow2h.view(b, c, 2, h, w) + up_flow3[:, np.newaxis]
        flow2h = flow2h.view(bs, -1, h, w)  # b, 12*2, h, w
        ent2h = ent2h.view(bs, -1, h, w)  # b, 12*1, h, w
        flow_reg_maxs.append(self.flow_reg4.flowx.max())

        # append coarse hypotheses
        flow2h = torch.cat(
            (
                flow2h,
                F.interpolate(
                    flow3h.detach() * 2,
                    [flow2h.shape[2], flow2h.shape[3]],
                    mode="bilinear",
                    align_corners=False,
                ),
            ),
            1,
        )
        ent2h = torch.cat(
            (
                ent2h,
                F.interpolate(
                    ent3h,
                    [flow2h.shape[2], flow2h.shape[3]],
                    mode="bilinear",
                    align_corners=False,
                ),
            ),
            1,
        )

        if self.training:
            x = torch.cat((ent2h.detach(), flow2h.detach(), c12), 1)
            oor2 = self.dc2_convo(x)[:, 0]
            oors.append(oor2)

        # hypotheses fusion
        x = torch.cat((ent2h.detach(), flow2h.detach(), c12), 1)
        x = self.dc2_conv4(self.dc2_conv3(self.dc2_conv2(self.dc2_conv1(x))))
        va = self.dc2_conv7(self.dc2_conv6(self.dc2_conv5(x)))
        va = va.view(b, -1, 2, h, w)
        flow2 = (flow2h.view(b, -1, 2, h, w) * F.softmax(va, 1)).sum(1)

        flow2 = F.interpolate(
            flow2, [im.size()[2], im.size()[3]], mode="bilinear", align_corners=False
        )
        flow2 = self.postprocess_predictions(flow2, image_resizer, is_flow=True)
        flow3 = F.interpolate(
            flow3, [im.size()[2], im.size()[3]], mode="bilinear", align_corners=False
        )
        flow3 = self.postprocess_predictions(flow3, image_resizer, is_flow=True)
        flow4 = F.interpolate(
            flow4, [im.size()[2], im.size()[3]], mode="bilinear", align_corners=False
        )
        flow4 = self.postprocess_predictions(flow4, image_resizer, is_flow=True)
        flow5 = F.interpolate(
            flow5, [im.size()[2], im.size()[3]], mode="bilinear", align_corners=False
        )
        flow5 = self.postprocess_predictions(flow5, image_resizer, is_flow=True)
        flow6 = F.interpolate(
            flow6, [im.size()[2], im.size()[3]], mode="bilinear", align_corners=False
        )
        flow6 = self.postprocess_predictions(flow6, image_resizer, is_flow=True)
        flows = [flow6, flow5, flow4, flow3, flow2]

        outputs = {}

        if self.training:
            outputs["flow_preds"] = flows
            outputs["up_flows_preds"] = up_flows
            outputs["oors_preds"] = oors
            outputs["flow_reg_maxs_preds"] = flow_reg_maxs
            outputs["flows"] = flow2[:, None] * 4
        else:
            outputs["flows"] = flow2[:, None] * 4

        return outputs


@register_model
@trainable
class vcn(VCN):
    pass


@register_model
@trainable
class vcn_small(VCNSmall):
    pass
